###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This is the main application of vistrail, it will calls for
initializations to the theme, packages and the builder...

"""
from __future__ import division

from ast import literal_eval
import os.path
import getpass
import re
import sys
import StringIO

from PyQt4 import QtGui, QtCore, QtNetwork

from vistrails.core.application import VistrailsApplicationInterface, \
    get_vistrails_application, set_vistrails_application
from vistrails.core import debug
from vistrails.core import system
from vistrails.core.application import APP_SUCCESS, APP_FAIL, APP_DONE
from vistrails.core.db.io import load_vistrail
from vistrails.core.db.locator import FileLocator, DBLocator
import vistrails.core.requirements
from vistrails.core.vistrail.controller import VistrailController
from vistrails.db import VistrailsDBException
from vistrails.db.services.io import test_db_connection
from vistrails.gui import qt
import vistrails.gui.theme


################################################################################

class VistrailsApplicationSingleton(VistrailsApplicationInterface,
                                    QtGui.QApplication):
    """
    VistrailsApplicationSingleton is the singleton of the application,
    there will be only one instance of the application during VisTrails
    
    """

    use_event_filter = system.systemType in ['Darwin']

    def __call__(self):
        """ __call__() -> VistrailsApplicationSingleton
        Return self for calling method
        
        """
        if not self._initialized and not self._is_running:
            self.init()
        return self

    def __init__(self):
        # font bugfix for Qt 4.8 and OS X 10.9
        import platform
        if platform.system()=='Darwin':
            release = platform.mac_ver()[0].split('.')
            if len(release)>=2 and int(release[0])*100+int(release[1])>=1009:
                QtGui.QFont.insertSubstitution(".Lucida Grande UI", "Lucida Grande")
        QtGui.QApplication.__init__(self, sys.argv)
        VistrailsApplicationInterface.__init__(self)

        if self.use_event_filter:
            self.installEventFilter(self)
        self.builderWindow = None
        # local notifications
        self.window_notifications = {}
        self.view_notifications = {}

        if QtCore.QT_VERSION < 0x40200: # 0x40200 = 4.2.0
            raise vistrails.core.requirements.MissingRequirement("Qt version >= 4.2")
        self._is_running = False
        self.shared_memory = None
        self.local_server = None
        self.setAttribute(QtCore.Qt.AA_DontShowIconsInMenus)
        qt.allowQObjects()

    def run_single_instance(self, args):
        # code for single instance of the application
        # based on the C++ solution available at
        # http://wiki.qtcentre.org/index.php?title=SingleApplication
        if QtCore.QT_VERSION >= 0x40400:
            self.timeout = 600000
            self._unique_key = os.path.join(system.home_directory(),
                       "vistrails-single-instance-check-%s"%getpass.getuser())
            self.shared_memory = QtCore.QSharedMemory(self._unique_key)
            self.local_server = None
            if self.shared_memory.attach():
                self._is_running = True

                local_socket = QtNetwork.QLocalSocket(self)
                local_socket.connectToServer(self._unique_key)
                if not local_socket.waitForConnected(self.timeout):
                    debug.critical(
                            "Connection failed: %s\n"
                            "Removing socket" % (local_socket.errorString()))
                    try:
                        os.remove(self._unique_key)
                    except OSError, e:
                        debug.critical("Couldn't remove socket: %s" %
                                       self._unique_key, e)

                else:
                    if self.found_another_instance_running(local_socket, args):
                        return APP_DONE # success, we should shut down
                    else:
                        return APP_FAIL  # error, we should shut down

            if not self.shared_memory.create(1):
                debug.critical("Unable to create single instance "
                               "of vistrails application")
                return
            self.local_server = QtNetwork.QLocalServer(self)
            self.connect(self.local_server, QtCore.SIGNAL("newConnection()"),
                         self.message_received)
            if self.local_server.listen(self._unique_key):
                debug.log("Listening on %s"%self.local_server.fullServerName())
            else:
                # This usually happens when vistrails have crashed
                # Delete the key and try again
                self.shared_memory.detach()
                self.local_server.close()
                if os.path.exists(self._unique_key):
                    os.remove(self._unique_key)

                self.shared_memory = QtCore.QSharedMemory(self._unique_key)
                self.local_server = None

                if not self.shared_memory.create(1):
                    debug.critical("Unable to create single instance "
                                   "of vistrails application")
                    return
                self.local_server = QtNetwork.QLocalServer(self)
                self.connect(self.local_server, QtCore.SIGNAL("newConnection()"),
                             self.message_received)
                if self.local_server.listen(self._unique_key):
                    debug.log("Listening on %s"%self.local_server.fullServerName())
                else:
                    debug.warning(
                            "Server is not listening. This means it "
                            "will not accept parameters from other "
                            "instances")

        return None

    def found_another_instance_running(self, local_socket, args):
        debug.critical("Found another instance of VisTrails running")
        msg = bytes(args)
        debug.critical("Will send parameters to main instance %s" % msg)
        res = self.send_message(local_socket, msg)
        if res is True:
            debug.critical("Main instance succeeded")
            return True
        elif res is False:
            return False
        else:
            debug.critical("Main instance reports: %s" % res)
            return False

    def init(self, optionsDict=None, args=[]):
        """ VistrailsApplicationSingleton(optionDict: dict)
                                          -> VistrailsApplicationSingleton
        Create the application with a dict of settings
        
        """
        vistrails.gui.theme.initializeCurrentTheme()
        VistrailsApplicationInterface.init(self, optionsDict, args)
        
        if self.temp_configuration.check('jobInfo') or \
           self.temp_configuration.check('jobList'):
            self.temp_configuration.batch = True

        # singleInstance configuration
        singleInstance = self.temp_configuration.check('singleInstance')
        if singleInstance:
            finished = self.run_single_instance(args)
            if finished is not None:
                return finished

        interactive = not self.temp_configuration.check('batch')
        if interactive:
            self.setIcon()
            self.createWindows()
            self.processEvents()
            
        # self.vistrailsStartup.init()
        self.package_manager.initialize_packages(
                report_missing_dependencies=not self.startup.first_run)

        # ugly workaround for configuration initialization order issue
        # If we go through the configuration too late,
        # The window does not get maximized. If we do it too early,
        # there are no created windows during spreadsheet initialization.
        if interactive:
            if  self.temp_configuration.check('maximizeWindows'):
                self.builderWindow.showMaximized()
            if self.temp_configuration.check('dbDefault'):
                self.builderWindow.setDBDefault(True)

        self._initialized = True

        # default handler installation
        if system.systemType == 'Linux':
            if not (self.temp_configuration.check('handlerDontAsk') or
                    self.configuration.check('handlerDontAsk')):
                if not linux_default_application_set():
                    self.ask_update_default_application()

        if interactive:
            self.interactiveMode()
        else:
            r = self.noninteractiveMode()
            return APP_SUCCESS if r is True else APP_FAIL
        return APP_SUCCESS

    def ask_update_default_application(self, dont_ask_checkbox=True):
        if hasattr(self, 'splashScreen') and self.splashScreen:
            self.splashScreen.hide()
        dialog = QtGui.QDialog()
        dialog.setWindowTitle(u"Install .vt .vtl handler")
        layout = QtGui.QVBoxLayout()
        dialog.setLayout(layout)
        layout.addWidget(QtGui.QLabel(u"Install VisTrails as default handler "
                                      u"to open .vt and .vtl files?"))
        if dont_ask_checkbox:
            dont_ask = QtGui.QCheckBox(u"Don't ask on startup")
            dont_ask_setting = self.configuration.check('handlerDontAsk')
            dont_ask.setChecked(dont_ask_setting)
            layout.addWidget(dont_ask)
        buttons = QtGui.QDialogButtonBox(
                QtGui.QDialogButtonBox.Yes | QtGui.QDialogButtonBox.No)
        layout.addWidget(buttons)
        QtCore.QObject.connect(buttons, QtCore.SIGNAL('accepted()'),
                     dialog, QtCore.SLOT('accept()'))
        QtCore.QObject.connect(buttons, QtCore.SIGNAL('rejected()'),
                     dialog, QtCore.SLOT('reject()'))

        res = dialog.exec_()
        if dont_ask_checkbox:
            if dont_ask.isChecked() != dont_ask_setting:
                self.configuration.handlerDontAsk = dont_ask.isChecked()
                self.configuration.handlerDontAsk = dont_ask.isChecked()
        if res != QtGui.QDialog.Accepted:
            return False
        if system.systemType == 'Linux':
            if not linux_update_default_application():
                QtGui.QMessageBox.warning(
                        None,
                        u"Install .vt .vtl handler",
                        u"Couldn't set VisTrails as default handler "
                        u"to open .vt and .vtl files")
                return False
        else:
            QtGui.QMessageBox.warning(
                    None,
                    u"Install .vt .vtl handler",
                    u"Can't install a default handler on this platform")
            return False
        return True

    def is_running_gui(self):
        return True

    def get_current_controller(self):
        return self.builderWindow.get_current_controller()
    get_controller = get_current_controller

    def get_vistrail(self):
        if self.get_controller():
            return self.get_controller().vistrail
        return None

    def ensure_vistrail(self, locator):
        view = self.builderWindow.ensureVistrail(locator)
        if view is not None:
            return view.controller
        return None

    def add_vistrail(self, *objs):
        return self.builderWindow.add_vistrail(*objs)

    def remove_vistrail(self, locator=None):
        return self.builderWindow.remove_vistrail(locator)

    def select_version(self, version):
        return self.builderWindow.select_version(version)

    def update_locator(self, old_locator, new_locator):
        pass

    def create_notification(self, notification_id, window=None, view=None):
        if view is not None:
            if view not in self.view_notifications:
                self.view_notifications[view] = {}
            notifications = self.view_notifications[view]
        elif window is not None:
            if window not in self.window_notifications:
                self.window_notifications[window] = {}
            notifications = self.window_notifications[window]
        else:
            notifications = self.notifications
        if notification_id not in notifications:
            notifications[notification_id] = set()
        # else:
        #     print "already added notification", notification_id

    def register_notification(self, notification_id, method, window=None,
                              view=None):
        if view is not None:
            if view not in self.view_notifications:
                self.view_notifications[view] = {}
            notifications = self.view_notifications[view]
            #print '>>> LOCAL adding notification', notification_id, view, method
            #print id(notifications), notifications
            #for n, o in notifications.iteritems():
            #    print "    ", n , "(%s)"%len(o)
            #    for m in o:
            #        print "        ", m
        elif window is not None:
            if window not in self.window_notifications:
                self.window_notifications[window] = {}
            notifications = self.window_notifications[window]
        else:
            notifications = self.notifications     
            #print '>>> GLOBAL adding notification', notification_id, method  
            #print id(notifications), notifications
        if notification_id not in notifications:
            self.create_notification(notification_id, window, view)
        notifications[notification_id].add(method)

    def unregister_notification(self, notification_id, method, window=None,
                                view=None):
        if view is not None:
            if view in self.view_notifications:
                notifications = self.view_notifications[view]
            else:
                notifications = {}
                #print '>>> LOCAL remove notification', notification_id, view
            
            #print id(notifications), notifications
#            for n, o in notifications.iteritems():
#                print "    ", n , "(%s)"%len(o)
#                for m in o:
#                    print "        ", m
        elif window is not None:
            if window in self.window_notifications:
                notifications = self.window_notifications[window]
            else:
                notifications = {}
        else:
            notifications = self.notifications    
            #print '>>> GLOBAL remove notification', notification_id, method   
            #print id(notifications), notifications           
        if notification_id in notifications:
            notifications[notification_id].remove(method)

    def send_notification(self, notification_id, *args):
        # do global notifications
        if notification_id in self.notifications:
            for m in self.notifications[notification_id]:
                try:
                    m(*args)
                except Exception, e:
                    debug.unexpected_exception(e)
                    debug.print_exc()
        notifications = {}

        current_window = self.builderWindow

        # do window notifications
        if current_window in self.window_notifications:
            notifications = self.window_notifications[current_window]

            if notification_id in notifications:
                for m in notifications[notification_id]:
                    try:
                        m(*args)
                    except Exception, e:
                        debug.unexpected_exception(e)
                        debug.print_exc()

        if current_window is not None:
            current_view = current_window.current_view
        else:
            current_view = None
        # do local notifications
        if current_view in self.view_notifications:
            notifications = self.view_notifications[current_view]

            if notification_id in notifications:
                for m in notifications[notification_id]:
                    try:
                        m(*args)
                    except Exception, e:
                        debug.unexpected_exception(e)
                        debug.print_exc()

    def showBuilderWindow(self):
        # in some systems (Linux and Tiger) we need to make both calls
        # so builderWindow is activated
        self.setActiveWindow(self.builderWindow)
        self.builderWindow.activateWindow()
        self.builderWindow.show()
        self.builderWindow.raise_()
    
    def interactiveMode(self):
        """ interactiveMode() -> None
        Instantiate the GUI for interactive mode
        
        """
        if self.temp_configuration.check('showSplash'):
            self.splashScreen.finish(self.builderWindow)
            debug.DebugPrint.getInstance().register_splash(None)
            self.splashScreen = None
        # self.builderWindow.modulePalette.updateFromModuleRegistry()
        # self.builderWindow.modulePalette.connect_registry_signals()
        self.builderWindow.link_registry()
        
        self.process_interactive_input()
        if self.temp_configuration.showWindow:
            self.showBuilderWindow()
        else:
            self.builderWindow.hide()
        self.builderWindow.create_first_vistrail()

    def noninteractiveMode(self):
        """ noninteractiveMode() -> None
        Run the console in non-interactive mode
        
        """
        usedb = False
        passwd = ''
        if self.temp_configuration.check('host'):
            usedb = True
        if usedb and self.temp_configuration.check('user'):
            db_config = dict((x, self.temp_configuration.check(x))
                             for x in ['host', 'port', 
                                       'db', 'user'])
            try:
                test_db_connection(db_config)
            except VistrailsDBException:
                passwd = \
                    getpass.getpass("Connecting to %s:%s. Password for user '%s':" % (
                                    db_config['host'],
                                    db_config['db'],
                                    db_config['user']))
                db_config['passwd'] = passwd
                try:
                    test_db_connection(db_config)
                except VistrailsDBException:
                    debug.critical("Cannot login to database")
                    return False
    
        if self.input:
            w_list = []
            vt_list = []
            for filename in self.input:
                f_name, version = self._parse_vtinfo(filename, not usedb)
                if not f_name:
                    debug.critical("File not found: %s" % filename)
                    return False
                if not usedb:
                    locator = FileLocator(os.path.abspath(f_name))
                else:
                    locator = DBLocator(
                           host=self.temp_configuration.check('host'),
                           port=self.temp_configuration.check('port') or 3306,
                           database=self.temp_configuration.check('db'),
                           user=self.temp_configuration.check('user'),
                           passwd=passwd,
                           obj_id=f_name,
                           obj_type=None,
                           connection_id=None)
                    if not locator.is_valid():
                        #here there is a problem: as we allow execution from 
                        #command line with VisTrails already running, we need
                        #to update from the gui
                        if hasattr(self, 'builderWindow'):
                            ok = locator.update_from_gui(self.builderWindow)
                        else:
                            ok = locator.update_from_console()
                        if not ok:
                            debug.critical("Cannot login to database")
                if self.temp_configuration.check('jobList'):
                    return self.printJobs(locator)
                if self.temp_configuration.check('jobInfo'):
                    return self.printJob(locator, version)

                w_list.append((locator, version))
                vt_list.append(locator)
            import vistrails.core.console_mode

            errs = []
            if self.temp_configuration.check('workflowGraph'):
                workflow_graph = self.temp_configuration.workflowGraph
                results = vistrails.core.console_mode.get_wf_graph(w_list, workflow_graph,
                                                                   self.temp_configuration.graphsAsPdf)
                for r in results:
                    if r[0] is False:
                        errs.append("Error generating workflow graph: %s" % \
                                    r[1])
                        debug.critical("*** Error in get_wf_graph: %s" % r[1])
            
            if self.temp_configuration.check('evolutionGraph'):
                evolution_graph = self.temp_configuration.evolutionGraph
                results = vistrails.core.console_mode.get_vt_graph(vt_list, evolution_graph,
                                                                   self.temp_configuration.graphsAsPdf)
                for r in results:
                    if r[0] is False:
                        errs.append("Error generating vistrail graph: %s" % \
                                    r[1])
                        debug.critical("*** Error in get_vt_graph: %s" % r[1])
                
            if self.temp_configuration.check('outputDirectory'):
                output_dir = self.temp_configuration.outputDirectory
            else:
                output_dir = None

            extra_info = None
            if self.temp_configuration.check('outputDirectory'):
                extra_info = \
                {'pathDumpCells': self.temp_configuration.outputDirectory}
            if self.temp_configuration.check('parameterExploration'):
                errs.extend(
                    vistrails.core.console_mode.run_parameter_explorations(
                        w_list, extra_info=extra_info))
            else:
                errs.extend(vistrails.core.console_mode.run(
                        w_list,
                        self.temp_configuration.check('parameters')
                            or '',
                        output_dir, update_vistrail=True,
                        extra_info=extra_info))
            if len(errs) > 0:
                for err in errs:
                    debug.critical("*** Error in %s:%s:%s -- %s" % err)
                return [False, ["*** Error in %s:%s:%s -- %s" % err for err in errs]]
            return True
        else:
            debug.warning("no input vistrails provided")
            return True

    def printJobs(self, locator):
        (v, abstractions, thumbnails, mashups)  = load_vistrail(locator)
        controller = VistrailController(v, locator, abstractions, thumbnails,
                                        mashups, auto_save=False)
        text = "### Workflows with jobs ###\n"
        text += "workflow | start date | status\n"
        text += '\n'.join(
            ["%s %s %s" %(j.version,
                          j.start,
                          "FINISHED" if j.completed() else "RUNNING")
             for i, j in controller.jobMonitor.workflows.iteritems()])
        print text
        return text

    def printJob(self, locator, version):
        (v, abstractions, thumbnails, mashups)  = load_vistrail(locator)
        controller = VistrailController(v, locator, abstractions, thumbnails,
                                        mashups, auto_save=False)
        text = "### Jobs in workflow ###\n"
        text += "name | start date | status\n"
        workflow = [wf for wf in controller.jobMonitor.workflows.itervalues()
                    if wf.version == int(version)]
        if len(workflow) < 1:
            text = "No job for workflow with id %s" % version
            print text
            return text
        workflow = workflow[0]
        text += '\n'.join(
            ["%s %s %s" %(i.name,
                          i.start,
                          "FINISHED" if i.finished else "RUNNING")
             for i in workflow.jobs.values()])
        print text
        return text

    def setIcon(self):
        """ setIcon() -> None
        Setup Vistrail Icon
        """
        self.setWindowIcon(vistrails.gui.theme.CurrentTheme.APPLICATION_ICON)
        
    def setupSplashScreen(self):
        """ setupSplashScreen() -> None
        Create the splash-screen at startup
        
        """
        if self.temp_configuration.check('showSplash'):
            splashPath = (system.vistrails_root_directory() +
                          "/gui/resources/images/vistrails_splash.png")
            pixmap = QtGui.QPixmap(splashPath)
            self.splashScreen = QtGui.QSplashScreen(pixmap, QtCore.Qt.WindowStaysOnTopHint)
            self.splashScreen.setFont(vistrails.gui.theme.CurrentTheme.SPLASH_SCREEN_FONT)
            debug.DebugPrint.getInstance().register_splash(self)
            self.splashScreen.show()
            
    def splashMessage(self, msg):
        if hasattr(self, "splashScreen"):
            self.splashScreen.showMessage(msg,
                        QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeft,
                        QtCore.Qt.white)
            self.processEvents()

    def createWindows(self):
        """ createWindows() -> None
        Create and configure all GUI widgets including the builder
        
        """
        self.setupSplashScreen()

        # This is so that we don't import too many things before we
        # have to. Otherwise, requirements are checked too late.
        # from gui.builder_window import QBuilderWindow
        from vistrails.gui.vistrails_window import QVistrailsWindow

        # self.builderWindow = QBuilderWindow()
        self.builderWindow = QVistrailsWindow()
        if self.temp_configuration.showWindow:
            # self.builderWindow.show()
            # self.setActiveWindow(self.builderWindow)
            pass

    def finishSession(self):
        if QtCore.QT_VERSION >= 0x40400 and self.shared_memory is not None:
            self.shared_memory.detach()
            if self.local_server:
                self.local_server.close()
        if system.systemType in ['Darwin']:
            self.removeEventFilter(self)
        VistrailsApplicationInterface.finishSession(self)
   
    def eventFilter(self, o, event):
        """eventFilter(obj,event)-> boolean
        This will filter all create events and will set on the WA_MacMetalStyle
        attribute of a QWidget. It will also filter the FileOpen events on Mac.
        
        """
        metalstyle = hasattr(self, 'temp_configuration') and \
                     self.temp_configuration.check('useMacBrushedMetalStyle')
        if metalstyle:
            if QtCore.QT_VERSION < 0x40500:
                create_event = QtCore.QEvent.Create
                mac_attribute = QtCore.Qt.WA_MacMetalStyle
            else:
                create_event = 15
                mac_attribute = QtCore.Qt.WA_MacBrushedMetal
            if (event.type() == create_event and
                    isinstance(o, QtGui.QWidget) and
                    not isinstance(o, QtGui.QSplashScreen) and
                    not (o.windowFlags() & QtCore.Qt.Popup)):
                o.setAttribute(mac_attribute)
        if event.type() == QtCore.QEvent.FileOpen:
            self.input = [str(event.file())]
            self.process_interactive_input()
        return QtGui.QApplication.eventFilter(self,o,event)
    
    def is_running(self):
        return self._is_running

    def message_received(self):
        if QtCore.QT_VERSION >= 0x40400:
            local_socket = self.local_server.nextPendingConnection()
            if not local_socket.waitForReadyRead(self.timeout):
                debug.critical("Read error: %s" %
                               local_socket.errorString())
                return
            byte_array = local_socket.readAll()
            self.temp_configuration.workflowGraph = None
            self.temp_configuration.evolutionGraph = None
            self.temp_configuration.outputDirectory = None
            self.temp_configuration.spreadsheetDumpPDF = False
            self.temp_configuration.execute = False
            self.temp_configuration.batch = False

            output = None
            try:
                # redirect stdout
                old_stdout = sys.stdout
                sys.stdout = StringIO.StringIO()
                result = self.parse_input_args_from_other_instance(str(byte_array))
                output = sys.stdout.getvalue()
                sys.stdout.close()
                sys.stdout = old_stdout
            except Exception, e:
                debug.unexpected_exception(e)
                debug.critical("Unknown error", e)
                result = debug.format_exc()
            if None == result:
                result = True
            if True == result:
                result = "Command Completed"
            elif False == result:
                result = "Command Failed"
            elif type(result) == list:
                result = '\n'.join(result[1])
            if result == "Command Completed" and output:
                result += '\n' + output
            self.shared_memory.lock()
            local_socket.write(bytes(result))
            self.shared_memory.unlock()
            if not local_socket.waitForBytesWritten(self.timeout):
                debug.critical("Writing failed: %s" %
                            local_socket.errorString())
                return
            local_socket.disconnectFromServer()

    def send_message(self, local_socket, message):
        self.shared_memory.lock()
        local_socket.write(message)
        self.shared_memory.unlock()
        if not local_socket.waitForBytesWritten(self.timeout):
            debug.critical("Writing failed: %s" %
                           local_socket.errorString())
            return False
        if not local_socket.waitForReadyRead(self.timeout):
            debug.critical("Read error: %s" %
                           local_socket.errorString())
            return False
        byte_array = local_socket.readAll()
        result = str(byte_array)
        print "Other instance processed input (%s)" % result
        if not result.startswith('Command Completed'):
            debug.critical(result)
        else:
            local_socket.disconnectFromServer()
            return True
        local_socket.disconnectFromServer()
        return False

    def parse_input_args_from_other_instance(self, msg):
        options_re = re.compile(r"^(\[('([^'])*', ?)*'([^']*)'\])|(\[\s?\])$")
        if options_re.match(msg):
            #it's safe to eval as a list
            args = literal_eval(msg)
            if isinstance(args, list):
                try:
                    self.read_options(args)
                except SystemExit:
                    debug.critical("Invalid options: %s" % ' '.join(args))
                    return False
                interactive = not self.temp_configuration.check('batch')
                if interactive:
                    result = self.process_interactive_input()
                    if self.temp_configuration.showWindow:
                        # in some systems (Linux and Tiger) we need to make both calls
                        # so builderWindow is activated
                        self.builderWindow.raise_()
                        self.builderWindow.activateWindow()
                    return result
                else:
                    return self.noninteractiveMode()
            else:
                debug.critical("Invalid string: %s" % msg)
        else:
            debug.critical("Invalid input: %s" % msg)
        return False

def linux_default_application_set():
    """linux_default_application_set() -> True|False|None
    For Linux - checks if a handler is set for .vt and .vtl files.
    """
    command = ['xdg-mime', 'query', 'filetype',
               os.path.join(system.vistrails_root_directory(),
                            'tests', 'resources', 'terminator.vt')]
    try:
        output = []
        result = system.execute_cmdline(command, output)
        if result != 0:
            # something is wrong, abort
            debug.warning("Error checking mimetypes: %s" % output[0])
            return None
    except OSError, e:
        debug.warning("Error checking mimetypes: %s" % e.message)
        return None
    if 'application/x-vistrails' == output[0].strip():
        return True
    return False

def linux_update_default_application():
    """ update_default_application() -> None
    For Linux - checks if we should install vistrails as the default
    application for .vt and .vtl files.
    If replace is False, don't replace an existing handler.

    Returns True if installation succeeded.
    """
    root = system.vistrails_root_directory()
    home = os.path.expanduser('~')

    # install mime type
    command = ['xdg-mime', 'install', 
               os.path.join(system.vistrails_root_directory(),
                            'gui/resources/vistrails-mime.xml')]
    output = []
    try:
        result = system.execute_cmdline(command, output)
    except OSError:
        result = None
    if result != 0:
        debug.warning("Error running xdg-mime")
        return False

    command = ['update-mime-database', home + '/.local/share/mime']
    output = []
    try:
        result = system.execute_cmdline(command, output)
    except OSError:
        result = None
    if result != 0:
        debug.warning("Error running update-mime-database")
        return False

    # install icon
    command = ['xdg-icon-resource', 'install',
               '--context', 'mimetypes',
               '--size', '48',
               os.path.join(system.vistrails_root_directory(),
                            'gui/resources/images/vistrails_icon_small.png'),
               'application-x-vistrails']
    output = []
    try:
        result = system.execute_cmdline(command, output)
    except OSError:
        result = None
    if result != 0:
        debug.warning("Error running xdg-icon-resource")
        return True # the handler is set anyway

    # install desktop file
    dirs = [home + '/.local', home + '/.local/share',
            home + '/.local/share/applications']

    for d in dirs:
        if not os.path.isdir(d):
            os.mkdir(d)
    desktop = """[Desktop Entry]
Name=VisTrails
Exec=python {root}/run.py %f
Icon={root}/gui/resources/images/vistrails_icon_small.png
Type=Application
MimeType=application/x-vistrails
""".format(root=root)
    f = open(os.path.join(dirs[2], 'vistrails.desktop'), 'w')
    f.write(desktop)
    f.close()

    command = ['update-desktop-database', dirs[2]]
    output = []
    try:
        result = system.execute_cmdline(command, output)
    except OSError:
        result = None
    if result != 0:
        debug.warning("Error running update-desktop-database")
    return True

# The initialization must be explicitly signalled. Otherwise, any
# modules importing vis_application will try to initialize the entire
# app.
def start_application(optionsDict=None, args=[]):
    """Initializes the application singleton."""
    VistrailsApplication = get_vistrails_application()
    if VistrailsApplication:
        debug.critical("Application already started.")
        return
    VistrailsApplication = VistrailsApplicationSingleton()
    set_vistrails_application(VistrailsApplication)
    x = VistrailsApplication.init(optionsDict, args)
    return x

def stop_application():
    """Stop and finalize the application singleton."""
    VistrailsApplication = get_vistrails_application()
    VistrailsApplication.finishSession()
    VistrailsApplication.save_configuration()
    VistrailsApplication.destroy()
    VistrailsApplication.deleteLater()




#!/usr/bin/env python

# This example demonstrates how to use the vtkPlaneWidget to probe
# a dataset and then generate contours on the probed data.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Start by loading some data.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()

# The plane widget is used probe the dataset.
planeWidget = vtk.vtkPlaneWidget()
planeWidget.SetInput(pl3d.GetOutput())
planeWidget.NormalToXAxisOn()
planeWidget.SetResolution(20)
planeWidget.SetRepresentationToOutline()
planeWidget.PlaceWidget()
plane = vtk.vtkPolyData()
planeWidget.GetPolyData(plane)

probe = vtk.vtkProbeFilter()
probe.SetInput(plane)
probe.SetSource(pl3d.GetOutput())

contourMapper = vtk.vtkPolyDataMapper()
contourMapper.SetInputConnection(probe.GetOutputPort())
contourMapper.SetScalarRange(pl3d.GetOutput().GetScalarRange())
contourActor = vtk.vtkActor()
contourActor.SetMapper(contourMapper)
contourActor.VisibilityOff()

# An outline is shown for context.
outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Actually generate contour lines.
def BeginInteraction(obj, event):
    global plane, contourActor
    obj.GetPolyData(plane)
    contourActor.VisibilityOn()

def ProbeData(obj, event):
    global plane
    obj.GetPolyData(plane)


# Associate the widget with the interactor
planeWidget.SetInteractor(iren)
# Handle the events.
planeWidget.AddObserver("EnableEvent", BeginInteraction)
planeWidget.AddObserver("StartInteractionEvent", BeginInteraction)
planeWidget.AddObserver("InteractionEvent", ProbeData)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(contourActor)

ren.SetBackground(1, 1, 1)
renWin.SetSize(300, 300)
ren.SetBackground(0.1, 0.2, 0.4)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 50)
cam1.SetFocalPoint(9.71821, 0.458166, 29.3999)
cam1.SetPosition(2.7439, -37.3196, 38.7167)
cam1.SetViewUp(-0.16123, 0.264271, 0.950876)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import math
import random
import copy

from itertools import imap, chain, izip

from vistrails.core.utils import all
from vistrails.core.data_structures.queue import Queue
from vistrails.core.data_structures.stack import Stack

import unittest
import random

################################################################################
# Graph

class GraphException(Exception):
    pass

class Graph(object):
    """Graph holds a graph with possible multiple edges. The
    datastructures are all dictionary-based, so datatypes more general than ints
    can be used. For example:
    
    >>> import graph
    >>> g = graph.Graph()
    >>> g.add_vertex('foo')
    >>> g.add_vertex('bar')
    >>> g.add_edge('foo', 'bar', 'edge_foo')
    >>> g.add_edge('foo', 'bar', 'edge_bar')
    >>> g.add_edge('bar', 'foo', 'edge_back')
    >>> g.out_degree('foo')
    2
    >>> g.out_degree('bar')
    1    
    """

    ##########################################################################
    # Constructor
    
    def __init__(self):
        """ Graph() -> Graph
        Initialize an empty graph and return nothing

        """
        self.vertices = {}
        self.adjacency_list = {}
        self.inverse_adjacency_list = {}

    @staticmethod
    def map_vertices(graph, vertex_map=None, edge_map=None):
        """ map_vertices(graph: Graph, vertex_map: dict): Graph

        Creates a new graph that is a mapping of vertex ids through
        vertex_map.

        """
        result = Graph()
        if vertex_map is None:
            vertex_map = dict((v, v) for v in graph.vertices)
        if edge_map is None:
            edge_map = {}
            for vfrom, lto in graph.adjacency_list.iteritems():
                for (vto, eid) in lto:
                    edge_map[eid] = eid
        result.vertices = dict((vertex_map[v], True) for v in graph.vertices)
        for (vto, lto) in graph.adjacency_list.iteritems():
            result.adjacency_list[vto] = [(vertex_map[to], edge_map[eid]) for (to, eid) in lto]
        for (vto, lto) in graph.inverse_adjacency_list.iteritems():
            result.inverse_adjacency_list[vto] = [(vertex_map[to], edge_map[eid]) for (to, eid) in lto]
        return result

    ##########################################################################
    # Accessors
            
    def inverse(self):
        """inverse() -> Graph
        Inverse all edge directions on the graph and return a Graph

        """
        result = copy.copy(self)
        t = result.adjacency_list
        result.adjacency_list = result.inverse_adjacency_list
        result.inverse_adjacency_list = t
        return result

    def inverse_immutable(self):
        """inverse_immutable() -> Graph
        
        Fast version of inverse(), but requires that output not be
        mutated (it shares with self.)
        """
        result = Graph()
        result.vertices = self.vertices
        result.adjacency_list = self.inverse_adjacency_list
        result.inverse_adjacency_list = self.adjacency_list
        return result

    def undirected_immutable(self):
        """undirected_immutable() -> Graph

        Creates an undirected version of self. Notice that this
        version should not be mutated because there is sharing in the
        adjacency lists and the vertex map.

        Additionally, if self wasn't acyclic, then
        undirected_immutable() won't be simple.
        """
        result = Graph()
        result.vertices = self.vertices
        result.adjacency_list = dict((k, (self.adjacency_list[k] +
                                          self.inverse_adjacency_list[k]))
                                     for k in self.vertices)
        result.inverse_adjacency_list = result.adjacency_list
        return result
        
    def out_degree(self, froom):
        """ out_degree(froom: id type) -> int
        Compute the number of edges leaving 'froom' and return an int

        Keyword arguments:
        froom -- 'immutable' vertex id

        """
        return len(self.adjacency_list[froom])
    
    def in_degree(self, to):
        """ in_degree(to: id type) -> int
        Compute the number of edges entering 'to' and return an int

        Keyword arguments:
        to -- 'immutable' vertex id

        """
        return len(self.inverse_adjacency_list[to])
    
    def sinks(self):
        """ sinks() -> list(id type)
        Find all vertices whose out_degree is zero and return a list of ids

        """
        return [idx for idx in self.vertices.keys() \
                if self.out_degree(idx) == 0]
    
    def sources(self):
        """ sources() -> list(id type)
        Find all vertices whose in_degree is zero and return a list of ids

        """
        return [idx for idx in self.vertices.keys() if self.in_degree(idx) == 0]

    def edges_to(self, id):
        """ edges_to(id: id type) -> list(list)
        Find edges entering a vertex id and return a list of tuples (id,id)

        Keyword arguments:
        id : 'immutable' vertex id
        
        """
        return self.inverse_adjacency_list[id]

    def edges_from(self, id):
        """ edges_from(id: id type) -> list(list)
        Find edges leaving a vertex id and return a list of tuples (id,id)
        
        Keyword arguments:
        id : 'immutable' vertex id

        """
        return self.adjacency_list[id]

    def get_edge(self, frm, to):
        """ get_edge(frm, to) -> edge_id

        Returns the id from the edge from->to."""
        for (t, e_id) in self.edges_from(frm):
            if t == to:
                return e_id

    def has_edge(self, frm, to):
        """ has_edge(frm, to) -> bool

        True if there exists an edge (frm, to)"""

        for (t, _) in self.edges_from(frm):
            if t == to:
                return True
        return False

    ##########################################################################
    # Mutate graph

    def add_vertex(self, id, data=None):
        """ add_vertex(id: id type, data) -> None
        Add a vertex to the graph if it is not already in the graph
        and return nothing

        Keyword arguments:
        id -- vertex id
        
        """
        if not id in self.vertices:
            self.vertices[id] = data
            self.adjacency_list[id] = []
            self.inverse_adjacency_list[id] = []

    def add_edge(self, froom, to, id=None):
        """ add_edge(froom: id type, to: id type, id: id type) -> None
        Add an edge from vertex 'froom' to vertex 'to' and return nothing

        Keyword arguments:
        froom -- 'immutable' origin vertex id
        to    -- 'immutable' destination vertex id
        id    -- 'immutable' edge id (default None)
          
        """
        self.add_vertex(froom)
        self.add_vertex(to)
        self.adjacency_list[froom].append((to, id))
        self.inverse_adjacency_list[to].append((froom, id))
        
    def delete_vertex(self, id):
        """ delete_vertex(id: id type) -> None
        Remove a vertex from graph and return nothing

        Keyword arguments:
        -- id : 'immutable' vertex id
          
        """
        
        for (origin, edge_id) in self.inverse_adjacency_list[id]:
            t = (id, edge_id)
            self.adjacency_list[origin].remove(t)
        for (dest, edge_id) in self.adjacency_list[id]:
            t = (id, edge_id)
            self.inverse_adjacency_list[dest].remove(t)
        del self.adjacency_list[id]
        del self.inverse_adjacency_list[id]
        del self.vertices[id]

    class RenameVertexError(GraphException):
        pass
    
    def rename_vertex(self, old_vertex, new_vertex):
        """ rename_vertex(old_vertex, new_vertex) -> None

        renames old_vertex to new_vertex in the graph, updating the edges
        appropriately. Should not be used to merge vertices, will raise
        exception if new_vertex exists in graph.

        """
        if not (old_vertex in self.vertices):
            raise self.RenameVertexError("vertex '%s' does not exist" % old_vertex) 
        if new_vertex in self.vertices:
            raise self.RenameVertexError("vertex '%s' already exists" % new_vertex)
        self.add_vertex(new_vertex)

        # the slice ([:]) is important for copying, since change_edge
        # mutates the list we'll be traversing
        for (v_from, e_id) in self.inverse_adjacency_list[old_vertex][:]:
            self.change_edge(v_from, old_vertex, new_vertex, e_id, e_id)

        self.adjacency_list[new_vertex] = self.adjacency_list[old_vertex]
        del self.adjacency_list[old_vertex]
        del self.vertices[old_vertex]
       
    def change_edge(self, old_froom, old_to, new_to, old_id=None, new_id=None):
        """ change_edge(old_froom: id, old_to: id, new_to: id, 
                        old_id: id, new_id: id) -> None
        Changes the destination of an edge in a graph **in place**
        
        Keyword arguments:
        old_froom -- 'immutable' origin vertex id
        old_to    -- 'immutable' destination vertex id
        new_to    -- 'immutable' destination vertex id
        old_id    -- 'immutable' edge id (default None)
        new_id    -- 'immutable' edge id (default None)
        """
        
        if old_id is None:
            efroom = self.adjacency_list[old_froom]
            forward_idx = None
            for i, edge in enumerate(efroom):
                if edge[0] == old_to:
                    old_id = edge[1]
                    forward_idx = i
                    break
            if forward_idx is None:
                raise ValueError("No edge to %r" % old_to)
        else:
            forward_idx = self.adjacency_list[old_froom].index((old_to, old_id))

        self.adjacency_list[old_froom][forward_idx] = ((new_to, new_id))
        self.inverse_adjacency_list[old_to].remove((old_froom, old_id))
        self.inverse_adjacency_list[new_to].append((old_froom, new_id))

    def delete_edge(self, froom, to, id=None):
        """ delete_edge(froom: id type, to: id type, id: id type) -> None
        Remove an edge from graph and return nothing

        Keyword arguments:
        froom -- 'immutable' origin vertex id
        to    -- 'immutable' destination vertex id
        id    -- 'immutable' edge id
          
        """
        if id is None:
            efroom = self.adjacency_list[froom]
            for edge in efroom:
                if edge[0] == to:
                    id = edge[1]
                    break
        if id is None:
            raise GraphException("delete_edge didn't find edge (%s,%s)"%
                                 (froom, to))
        self.adjacency_list[froom].remove((to, id))
        self.inverse_adjacency_list[to].remove((froom, id))

    ##########################################################################
    # Graph algorithms

    def closest_vertex(self, frm, target_list):
        """ closest_vertex(frm, target_list) -> id Uses bfs-like
        algorithm to find closest vertex to frm in target_list

        """
        if frm in target_list:
            return frm
        target_list = set(target_list)
        visited = set([frm])
        parent = {}
        q = Queue()
        q.push(frm)
        while 1:
            try:
                current = q.pop()
            except q.EmptyQueue:
                raise GraphException("no vertices reachable: %s %s" % (frm, list(target_list)))
            efrom = self.edges_from(current)
            for (to, eid) in efrom:
                if to in target_list:
                    return to
                if to not in visited:
                    parent[to] = current
                    q.push(to)
                    visited.add(to)

    def bfs(self, frm):
        """ bfs(frm:id type) -> dict(id type)
        Perform Breadth-First-Search and return a dict of parent id

        Keyword arguments:
        frm -- 'immutable' vertex id

        """
        visited = set([frm])
        parent = {}
        q = Queue()
        q.push(frm)
        while 1:
            try:
                current = q.pop()
            except q.EmptyQueue:
                break
            efrom = self.edges_from(current)
            for (to, eid) in efrom:
                if to not in visited:
                    parent[to] = current
                    q.push(to)
                    visited.add(to)
        return parent

    class GraphContainsCycles(GraphException):
        def __init__(self, v1, v2):
            self.back_edge = (v1, v2)
        def __str__(self):
            return ("Graph contains cycles: back edge %s encountered" %
                    self.back_edge)

    def dfs(self,
            vertex_set=None,
            raise_if_cyclic=False,
            enter_vertex=None,
            leave_vertex=None):
        """ dfs(self,vertex_set=None,raise_if_cyclic=False,enter_vertex=None,
                leave_vertex=None) -> (discovery, parent, finish)
        Performs a depth-first search on a graph and returns three dictionaries with
        relevant information. If vertex_set is not None, then it is used as
        the list of ids to perform the DFS on.
        
        See CLRS p. 541.

        enter_vertex, when present, is called just before visiting a vertex
        for the first time (and only once) with the vertex id as a parameter.

        leave_vertex, when present, is called just after visiting a vertex
        for the first time (and only once) with the vertex id as a parameter.
        """

        if not vertex_set:
            vertex_set = self.vertices

        # Ugly ugly python
        # http://mail.python.org/pipermail/python-list/2006-April/378964.html

        # We cannot explicitly "del data":
        # http://www.python.org/dev/peps/pep-0227/

        class Closure(object):
            
            def clear(self):
                del self.discovery
                del self.parent
                del self.finish
                del self.t
        
        # Straight CLRS p.541
        data = Closure()
        data.discovery = {} # d in CLRS
        data.parent = {} # \pi in CLRS
        data.finish = {}  # f in CLRS
        data.t = 0

        (enter, leave, back, other) = xrange(4)

        # inspired by http://www.ics.uci.edu/~eppstein/PADS/DFS.py

        def handle(v, w, edgetype):
            data.t += 1
            if edgetype == enter:
                data.discovery[v] = data.t
                if enter_vertex:
                    enter_vertex(w)
                if v != w:
                    data.parent[w] = v
            elif edgetype == leave:
                data.finish[w] = data.t
                if leave_vertex:
                    leave_vertex(w)
            elif edgetype == back and raise_if_cyclic:
                raise self.GraphContainsCycles(v, w)
        
        visited = set()
        gray = set()
        # helper function to build stack structure
        def st(v): return (v, iter(self.adjacency_list[v]))
        for vertex in vertex_set:
            if vertex not in visited:
                handle(vertex, vertex, enter)
                visited.add(vertex)
                stack = Stack()
                stack.push(st(vertex))
                gray.add(vertex)
                while stack.size:
                    parent, children = stack.top()
                    try:
                        child, _ = children.next()
                        if child in visited:
                            handle(parent, child, (child in gray
                                                   and back
                                                   or other))
                        else:
                            handle(parent, child, enter)
                            visited.add(child)
                            stack.push(st(child))
                            gray.add(child)
                    except StopIteration:
                        gray.remove(parent)
                        stack.pop()
                        if stack.size:
                            handle(stack.top()[0], parent, leave)
                handle(vertex, vertex, leave)

        result = (data.discovery, data.parent, data.finish)
        data.clear()
        return result

    class VertexHasNoParentError(GraphException):
        def __init__(self, v):
            Exception.__init__(self, v)
            self._v = v
        def __str__(self):
            return ("called parent() on vertex '%s', which has no parent nodes"
                    % self._v)

    def parent(self, v):
        """ parent(v: id type) -> id type
        Find the parent of vertex v and return an id

        Keyword arguments:
        v -- 'immutable' vertex id

        raises VertexHasNoParentError is vertex has no parent

        raises KeyError is vertex is not on graph

        """
        l=self.inverse_adjacency_list[v]
        if len(l):
            (froom, a) = l[-1]
        else:
            raise self.VertexHasNoParentError(v)
        return froom
    
    def vertices_topological_sort(self,vertex_set=None):
        """ vertices_topological_sort(self,vertex_set=None) ->
        sequence(vertices) Returns a sequence of all vertices, so that
        they are in topological sort order (every node traversed is
        such that their parent nodes have already been
        traversed). vertex_set is optionally a list of vertices on
        which to perform the topological sort.

        This is O(n log n) instead of the optimal O(n), 
        """
        (d, p, f) = self.dfs(vertex_set,raise_if_cyclic=True)
        # Optimized these three lines into the last one
        # lst = [(v, k) for (k,v) in f.iteritems()]
        # lst.sort(reverse=True)
        # return [v for (k, v) in lst]
        return [k for (k, _) in sorted(f.iteritems(), 
                                       key=lambda x: (x[1], x[0]),
                                       reverse=True)]

    def topologically_contractible(self, subgraph):
        """topologically_contractible(subgraph) -> Boolean.

        Returns true if contracting the subgraph to a single vertex
        doesn't create cycles. This is equivalent to checking whether
        a pipeline subgraph forms a legal abstraction."""
        x = copy.copy(self)
        conns_to_subgraph = self.connections_to_subgraph(subgraph)
        conns_from_subgraph = self.connections_from_subgraph(subgraph)
        for v in subgraph.vertices.iterkeys():
            x.delete_vertex(v)
        free_vertex = max(subgraph.vertices.iterkeys()) + 1
        x.add_vertex(free_vertex)
        for (edge_from, edge_to, edge_id) in conns_to_subgraph:
            x.add_edge(free_vertex, edge_to)
        for (edge_from, edge_to, edge_id) in conns_from_subgraph:
            x.add_edge(edge_from, free_vertex)
        try:
            x.vertices_topological_sort()
            return True
        except self.GraphContainsCycles:
            return False

    ##########################################################################
    # Subgraphs

    def subgraph(self, vertex_set):
        """ subgraph(vertex_set) -> Graph.

        Returns a subgraph of self containing all vertices and
        connections between them."""
        result = Graph()
        vertex_set = set(vertex_set)
        # add vertices
        for vertex in vertex_set:
            result.add_vertex(vertex)
        # add edges
        for vertex_from in vertex_set:
            for (vertex_to, edge_id) in self.edges_from(vertex_from):
                if vertex_to in vertex_set:
                    result.add_edge(vertex_from, vertex_to, edge_id)
        return result

    def connections_to_subgraph(self, subgraph):
        """connections_to_subgraph(subgraph) -> [(vert_from, vert_to, edge_id)]

        Returns the list of all edges that connect to a vertex \in
        subgraph. subgraph is assumed to be a subgraph of self"""
        vertices_to_traverse = set(self.vertices.iterkeys())
        subgraph_verts = set(subgraph.vertices.iterkeys())
        vertices_to_traverse -= subgraph_verts

        result = []
        for v in vertices_to_traverse:
            for e in self.adjacency_list[v]:
                (v_to, e_id) = e
                if v_to in subgraph_verts:
                    result.append((v, v_to, e_id))
        return result

    def connections_from_subgraph(self, subgraph):
        """connections_from_subgraph(subgraph) -> [(vert_from, vert_to, edge_id)]

        Returns the list of all edges that connect from a vertex \in
        subgraph to a vertex \not \in subgraph. subgraph is assumed to
        be a subgraph of self"""
        subgraph_verts = set(subgraph.vertices.iterkeys())
        vertices_to_traverse = subgraph_verts

        result = []
        for v in vertices_to_traverse:
            for e in self.adjacency_list[v]:
                (v_to, e_id) = e
                if v_to not in subgraph_verts:
                    result.append((v, v_to, e_id))
        return result

    ##########################################################################
    # Iterators


    def iter_edges_from(self, vertex):
        """iter_edges_from(self, vertex) -> iterable

        Returns an iterator over all edges in the form
        (vertex, vert_to, edge_id)."""
        def fn(edge):
            (edge_to, edge_id) = edge
            return (vertex, edge_to, edge_id)
        return imap(fn, self.adjacency_list[vertex])

    def iter_edges_to(self, vertex):
        """iter_edges_to(self, vertex) -> iterable

        Returns an iterator over all edges in the form
        (vertex, vert_to, edge_id)."""
        def fn(edge):
            (edge_from, edge_id) = edge
            return (edge_from, vertex, edge_id)
        return imap(fn, self.inverse_adjacency_list[vertex])

    def iter_all_edges(self):
        """iter_all_edges() -> iterable

        Returns an iterator over all edges in the graph in the form
        (vert_from, vert_to, edge_id)."""
        verts = self.iter_vertices()
        edge_itors = imap(self.iter_edges_from, verts)
        return chain(*[v for v in edge_itors])

    def iter_vertices(self):
        """iter_vertices() -> iterable

        Returns an iterator over all vertex ids of the graph."""
        return self.vertices.iterkeys()

    ##########################################################################
    # Special Python methods

    def __str__(self):
        """ __str__() -> str
        Format the graph for serialization and return a string

        """
        vs = self.vertices.keys()
        vs.sort()
        al = [(vfrom, vto, edgeid)
              for vfrom, lto in self.adjacency_list.iteritems()
              for vto, edgeid in lto]
        al.sort()
        return ("digraph G {\n"
                + ";".join([str(s) for s in vs])
                + ";\n"
                + "\n".join(["%s -> %s [label=\"%s\"];" % s for s in al])
                + "\n}")

    def __repr__(self):
        """ __repr__() -> str
        Similar to __str__ to re-represent the graph and returns a string

        """
        return self.__str__()

    def __copy__(self):
        """ __copy__() -> Graph
        Make a copy of the graph and return a Graph

        """
        cp = Graph()
        cp.vertices = copy.copy(self.vertices)
        cp.adjacency_list = dict((k, v[:]) for (k,v) in self.adjacency_list.iteritems())
        cp.inverse_adjacency_list = dict((k, v[:]) for (k,v) in self.inverse_adjacency_list.iteritems())
        return cp

    def __eq__(self, other):
        # Does not test isomorphism - vertices must be consistently labeled
        # might be slow - don't use in tight code
        if type(self) <> type(other):
            return False
        for v in self.vertices:
            if not v in other.vertices:
                return False
        for vfrom, elist in self.adjacency_list.iteritems():
            for vto, eid in elist:
                if not other.get_edge(vfrom, vto) == eid:
                    return False
        return True

    def __ne__(self, other):
        return not (self == other)

################################################################################
# Unit testing


class TestGraph(unittest.TestCase):
    """ Class to test Graph

    It tests vertex addition, the out_degree of a sink and in_degree of a
    source consistencies.
    
    """

    def make_complete(self, v):
        """returns a complete graph with v verts."""
        g = Graph()
        for x in xrange(v):
            g.add_vertex(x)
        for f in xrange(v):
            for t in xrange(f+1, v):
                g.add_edge(f, t, f * v + t)
        return g

    def make_linear(self, v, bw=False):
        """returns a linear graph with v verts. if bw=True, add
        backward links."""
        g = Graph()
        for x in xrange(v):
            g.add_vertex(x)
        for x,y in izip(xrange(v-1), xrange(1, v)):
            g.add_edge(x, y, x)
            if bw:
                g.add_edge(y, x, x + v)
        return g

    def get_default_graph(self):
        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        g.add_vertex(2)
        g.add_vertex(3)
        g.add_vertex(4)
        g.add_edge(0,1,0)
        g.add_edge(1,2,1)
        g.add_edge(0,3,2)
        g.add_edge(3,2,3)
        g.add_edge(2,4,4)
        return g
    
    def test1(self):
        """Test adding edges and vertices"""
        g = Graph()
        g.add_vertex('0')
        g.add_vertex('1')
        g.add_vertex('2')
        g.add_vertex('3')
        g.add_edge('0', '1', 0)
        g.add_edge('1', '2', 1)
        g.add_edge('2', '3', 2)
        parent = g.bfs('0')
        self.assertEquals(parent['3'], '2')
        self.assertEquals(parent['2'], '1')
        self.assertEquals(parent['1'], '0')

    def test2(self):
        """Test bread-first-search"""
        g = self.get_default_graph()
        p = g.bfs(0)
        k = p.keys()
        k.sort()
        self.assertEquals(k, [1, 2, 3, 4])
        inv = g.inverse()
        p_inv = inv.bfs(4)
        k2 = p_inv.keys()
        k2.sort()
        self.assertEquals(k2, [0, 1, 2, 3])
        
    def test3(self):
        """Test sink and source degree consistency"""
        g = Graph()
        for i in xrange(100):
            g.add_vertex(i)
        for i in xrange(1000):
            v1 = random.randint(0,99)
            v2 = random.randint(0,99)
            g.add_edge(v1, v2, i)
        sinkResult = [None for i in g.sinks() if g.out_degree(i) == 0]
        sourceResult = [None for i in g.sources() if g.in_degree(i) == 0]
        if len(sinkResult) <> len(g.sinks()):
            assert False
        if len(sourceResult) <> len(g.sources()):
            assert False

    def test_remove_vertices(self):
        g = self.make_linear(5)
        g.delete_vertex(1)
        g.delete_vertex(2)
    
    def test_DFS(self):
        """Test DFS on graph."""
        g = self.get_default_graph()
        g.dfs()

    def test_topological_sort(self):
        """Test toposort on graph."""
        g = self.get_default_graph()
        g.vertices_topological_sort()

        g = self.make_linear(10)
        r = g.vertices_topological_sort()
        assert r == [0,1,2,3,4,5,6,7,8,9]

        g = Graph()
        g.add_vertex('a')
        g.add_vertex('b')
        g.add_vertex('c')
        g.add_edge('a', 'b')
        g.add_edge('b', 'c')
        assert g.vertices_topological_sort() == ['a', 'b', 'c']

    def test_limited_DFS(self):
        """Test DFS on graph using a limited set of starting vertices."""
        g = self.get_default_graph()
        g.dfs(vertex_set=[1])
        g.dfs(vertex_set=[1,3])
        g.dfs(vertex_set=[1,2])

    def test_limited_topological_sort(self):
        """Test toposort on graph using a limited set of starting vertices."""
        g = self.get_default_graph()
        g.vertices_topological_sort(vertex_set=[1])
        g.vertices_topological_sort(vertex_set=[1,3])
        g.vertices_topological_sort(vertex_set=[1,2])

    def test_print_empty_graph(self):
        """Test print on empty graph"""
        g = Graph()
        g.__str__()

    def test_delete(self):
        """Tests consistency of data structure after deletion."""
        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        g.add_vertex(2)
        g.add_edge(0, 1, 0)
        g.add_edge(1, 2, 1)
        g.delete_vertex(2)
        self.assertEquals(g.adjacency_list[1], [])

    def test_raising_DFS(self):
        """Tests if DFS with cycle-checking will raise exceptions."""
        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        g.add_vertex(2)
        g.add_edge(0, 1)
        g.add_edge(1, 2)
        g.add_edge(2, 0)
        with self.assertRaises(Graph.GraphContainsCycles):
            g.dfs(raise_if_cyclic=True)

    def test_call_inverse(self):
        """Test if calling inverse methods work."""
        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        g.add_vertex(2)
        g.add_edge(0, 1)
        g.add_edge(1, 2)
        g.add_edge(2, 0)
        g2 = g.inverse()
        g3 = g.inverse_immutable()
    
    def test_subgraph(self):
        """Test subgraph routines."""
        g = self.make_complete(5)
        sub = g.subgraph([0,1])
        assert 0 in sub.vertices
        assert 1 in sub.vertices
        assert (1,1) in sub.adjacency_list[0]
        assert (0,1) in sub.inverse_adjacency_list[1]

        g = self.make_linear(3)
        sub = g.subgraph([0, 2])
        assert 0 in sub.vertices
        assert 2 in sub.vertices
        assert sub.adjacency_list[0] == []
        assert sub.adjacency_list[2] == []
        
    def test_connections_to_subgraph(self):
        """Test connections_to_subgraph."""
        g = self.make_linear(5)
        sub = g.subgraph([3])
        assert len(g.connections_to_subgraph(sub)) == 1
        g = self.make_linear(5, True)
        sub = g.subgraph([3])
        assert len(g.connections_to_subgraph(sub)) == 2

    def test_connections_from_subgraph(self):
        """Test connections_from_subgraph."""
        g = self.make_linear(5)
        sub = g.subgraph([3])
        assert len(g.connections_from_subgraph(sub)) == 1
        g = self.make_linear(5, True)
        sub = g.subgraph([3])
        assert len(g.connections_from_subgraph(sub)) == 2

    def test_topologically_contractible(self):
        """Test topologically_contractible."""
        g = self.make_linear(5)
        sub = g.subgraph([1, 2])
        assert g.topologically_contractible(sub)
        sub = g.subgraph([1, 3])
        assert not g.topologically_contractible(sub)

        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        g.add_vertex(2)
        g.add_vertex(3)
        g.add_edge(0, 1)
        g.add_edge(2, 3)
        for i in xrange(1, 16):
            s = []
            for j in xrange(4):
                if i & (1 << j): s.append(j)
            assert g.topologically_contractible(g.subgraph(s))

    def test_iter_vertices(self):
        g = self.get_default_graph()
        l = list(g.iter_vertices())
        l.sort()
        assert l == [0,1,2,3,4]

    def test_iter_edges(self):
        g = self.get_default_graph()
        l = [v for v in g.iter_all_edges()]
        l.sort()
        assert l == [(0,1,0), (0,3,2), (1, 2, 1), (2, 4, 4), (3, 2, 3)]

    def test_iter_edges_empty(self):
        """Test iterators on empty parts of the graph."""
        g = Graph()
        for a in g.iter_vertices():
            assert False
        g.add_vertex(0)
        for a in g.iter_edges_from(0):
            assert False
        for a in g.iter_edges_to(0):
            assert False
        for a in g.iter_all_edges():
            assert False

    def test_get_edge_none(self):
        g = Graph()
        g.add_vertex(0)
        g.add_vertex(1)
        assert g.get_edge(0, 1) is None

    def test_dfs_before(self):
        g = self.make_linear(10)
        inc = []
        dec = []
        def before(id): inc.append(id)
        def after(id): dec.append(id)
        g.dfs(enter_vertex=before,
              leave_vertex=after)
        assert inc == [0,1,2,3,4,5,6,7,8,9]
        assert inc == list(reversed(dec))
        assert all(a < b for a, b in izip(inc[:-1], inc[1:]))
        assert all(a > b for a, b in izip(dec[:-1], dec[1:]))

    def test_parent_source(self):
        g = self.make_linear(10)
        self.assertRaises(g.VertexHasNoParentError,
                          lambda: g.parent(0))
        for i in xrange(1, 10):
            assert g.parent(i) == i-1

    def test_rename_vertex(self):
        g = self.make_linear(10)
        self.assertRaises(g.RenameVertexError,
                          lambda: g.rename_vertex(0, 1))
        assert g.get_edge(0, 1) is not None
        assert g.get_edge(0, 11) is None
        g.rename_vertex(1, 11)
        assert g.get_edge(0, 1) is None
        assert g.get_edge(0, 11) is not None
        g.rename_vertex(11, 1)
        assert g.get_edge(0, 1) is not None
        assert g.get_edge(0, 11) is None

    def test_delete_get_edge(self):
        g = self.make_linear(10)
        self.assertRaises(GraphException, lambda: g.delete_edge(7, 9))
        assert g.has_edge(7, 8)
        g.delete_edge(7, 8)
        assert not g.has_edge(7, 8)

    def test_bfs(self):
        g = self.make_linear(5)
        lst = g.bfs(0).items()
        lst.sort()
        assert lst == [(1, 0), (2, 1), (3, 2), (4, 3)]
        lst = g.bfs(2).items()
        lst.sort()
        assert lst == [(3, 2), (4, 3)]

    def test_undirected(self):
        g = self.make_linear(5).undirected_immutable()
        lst = g.bfs(0).items()
        lst.sort()
        assert lst == [(1, 0), (2, 1), (3, 2), (4, 3)]
        lst = g.bfs(2).items()
        lst.sort()
        assert lst == [(0, 1), (1, 2), (3, 2), (4, 3)]

    def test_closest_vertex(self):
        g = self.make_linear(10)
        g.delete_edge(7, 8)
        g = g.undirected_immutable()
        self.assertRaises(GraphException, lambda: g.closest_vertex(1, [9]))
        assert g.closest_vertex(3, [2, 6, 7]) == 2
        assert g.closest_vertex(3, [2, 3, 6, 7]) == 3
        # Test using dictionary as target_list

        d1 = {2:True, 6:True, 7:False}
        d2 = {2:True, 6:True, 7:False, 3:False}
        d3 = {9:True}
        self.assertRaises(GraphException, lambda: g.closest_vertex(1, d3))
        assert g.closest_vertex(3, d1) == 2
        assert g.closest_vertex(3, d2) == 3

    def test_copy_not_share(self):
        g = self.make_linear(10)
        g2 = copy.copy(g)
        for v in g.vertices:
            assert id(g.adjacency_list[v]) <> id(g2.adjacency_list[v])
            assert id(g.inverse_adjacency_list[v]) <> id(g2.inverse_adjacency_list[v])

    def test_copy_works(self):
        g = self.make_linear(10)
        g2 = copy.copy(g)
        for v in g.vertices:
            assert v in g2.vertices
            assert g2.adjacency_list[v] == g.adjacency_list[v]
            assert g2.inverse_adjacency_list[v] == g.inverse_adjacency_list[v]

    def test_equals(self):
        g = self.make_linear(5)
        assert copy.copy(g) == g
        g2 = copy.copy(g)
        g2.add_vertex(10)
        assert g2 <> g

    def test_map_vertices(self):
        g = self.make_linear(5)
        m = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4}
        assert g == Graph.map_vertices(g, m)
        m = {0: 5, 1: 6, 2: 7, 3: 8, 4: 9}
        assert g <> Graph.map_vertices(g, m)
        
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# This example shows how to use decimation to reduce a polygonal
# mesh. We also use mesh smoothing and generate surface normals to
# give a pleasing result.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# We start by reading some data that was originally captured from a
# Cyberware laser digitizing system.
fran = vtk.vtkPolyDataReader()
fran.SetFileName(VTK_DATA_ROOT + "/Data/fran_cut.vtk")

# We want to preserve topology (not let any cracks form). This may
# limit the total reduction possible, which we have specified at 90%.
deci = vtk.vtkDecimatePro()
deci.SetInputConnection(fran.GetOutputPort())
deci.SetTargetReduction(0.9)
deci.PreserveTopologyOn()
smoother = vtk.vtkSmoothPolyDataFilter()
smoother.SetInputConnection(deci.GetOutputPort())
smoother.SetNumberOfIterations(50)
normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(smoother.GetOutputPort())
normals.FlipNormalsOn()
franMapper = vtk.vtkPolyDataMapper()
franMapper.SetInputConnection(normals.GetOutputPort())
franActor = vtk.vtkActor()
franActor.SetMapper(franMapper)
franActor.GetProperty().SetColor(1.0, 0.49, 0.25)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(franActor)
ren.SetBackground(1, 1, 1)
renWin.SetSize(250, 250)

cam1 = vtk.vtkCamera()
cam1.SetClippingRange(0.0475572, 2.37786)
cam1.SetFocalPoint(0.052665, -0.129454, -0.0573973)
cam1.SetPosition(0.327637, -0.116299, -0.256418)
cam1.SetViewUp(-0.0225386, 0.999137, 0.034901)
ren.SetActiveCamera(cam1)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""This file implements the Spreadsheet Tab Controller, to manages tabs:
StandardWidgetTabController
"""

from __future__ import division

from ast import literal_eval
import copy
import gc
import os.path
from PyQt4 import QtCore, QtGui

from vistrails.core.db.locator import FileLocator, _DBLocator as DBLocator
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.core.utils import DummyView
from vistrails.core.utils.uxml import XMLWrapper, named_elements
from vistrails.db.services.io import SaveBundle
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.utils import show_warning

from .spreadsheet_registry import spreadsheetRegistry
from .spreadsheet_tab import StandardWidgetTabBar, StandardWidgetSheetTab, \
    StandardTabDockWidget


class StandardWidgetTabController(QtGui.QTabWidget):
    """
    StandardWidgetTabController inherits from QTabWidget to contain a
    list of StandardWidgetSheetTab. This is the major component that
    will handle most of the spreadsheet actions

    """
    def __init__(self, parent=None):
        """ StandardWidgetTabController(parent: QWidget)
                                        -> StandardWidgetTabController
        Initialize signals/slots and widgets for the tab bar

        """

        QtGui.QTabWidget.__init__(self, parent)
        self.operatingWidget = self
        self.setTabBar(StandardWidgetTabBar(self))
        self.setTabShape(QtGui.QTabWidget.Triangular)
        self.setTabPosition(QtGui.QTabWidget.South)
        self.tabWidgets = []
        self.floatingTabWidgets = []
        self.addTabWidget(StandardWidgetSheetTab(self), 'Sheet 1')
        self.connect(self.tabBar(),
                     QtCore.SIGNAL('tabMoveRequest(int,int)'),
                     self.moveTab)
        self.connect(self.tabBar(),
                     QtCore.SIGNAL('tabSplitRequest(int,QPoint)'),
                     self.splitTab)
        self.connect(self.tabBar(),
                     QtCore.SIGNAL('tabTextChanged(int,QString)'),
                     self.changeTabText)
        self.addAction(self.showNextTabAction())
        self.addAction(self.showPrevTabAction())
        self.executedPipelines = [[],{},{}]
        self.monitoredPipelines = {}
        self.spreadsheetFileName = None
        self.loadingMode = False
        self.editingMode = False
        self.closeButton = QtGui.QToolButton(self)
        self.closeButton.setIcon(CurrentTheme.VIEW_MANAGER_CLOSE_ICON)
        self.closeButton.setAutoRaise(True)
        self.setCornerWidget(self.closeButton)
        self.connect(self.closeButton, QtCore.SIGNAL('clicked()'),
                     self.deleteSheetAction().trigger)

    def isLoadingMode(self):
        """ isLoadingMode() -> boolean
        Checking if the controller is in loading mode

        """
        return self.loadingMode

    def getMonitoredLocations(self, spec):
        """ getMonitoredLocations(spec: tuple) -> location
        Return the monitored location associated with spec

        """
        key = ((spec[0]['locator'], spec[0]['version']), spec[1], spec[2])
        if key in self.monitoredPipelines:
            return self.monitoredPipelines[key]
        else:
            return []

    def appendMonitoredLocations(self, spec, value):
        """ getMonitoredLocations(spec: tuple, value: location) -> None
        Return the monitored location associated with spec

        """
        key = ((spec[0]['locator'], spec[0]['version']), spec[1], spec[2])
        if key in self.monitoredPipelines:
            self.monitoredPipelines[key].append(value)
        else:
            self.monitoredPipelines[key] = [value]

    def newSheetAction(self):
        """ newSheetAction() -> QAction
        Return the 'New Sheet' action

        """
        if not hasattr(self, 'newSheetActionVar'):
            icon = QtGui.QIcon(':/images/newsheet.png')
            self.newSheetActionVar = QtGui.QAction(icon, '&New sheet', self)
            self.newSheetActionVar.setToolTip('Create a new sheet')
            self.newSheetActionVar.setStatusTip('Create and show a new sheet')
            self.newSheetActionVar.setShortcut(QtGui.QKeySequence('Ctrl+N'))
            self.connect(self.newSheetActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.newSheetActionTriggered)
        return self.newSheetActionVar

    def deleteSheetAction(self):
        """ deleteSheetAction() -> QAction
        Return the 'Delete Sheet' action:

        """
        if not hasattr(self, 'deleteSheetActionVar'):
            icon = QtGui.QIcon(':/images/deletesheet.png')
            self.deleteSheetActionVar = QtGui.QAction(icon, '&Delete sheet',
                                                      self)
            self.deleteSheetActionVar.setToolTip('Delete the current sheet')
            self.deleteSheetActionVar.setStatusTip('Delete the current sheet '
                                                   'if there are more than one')
            key = QtGui.QKeySequence('Ctrl+Backspace')
            self.deleteSheetActionVar.setShortcut(key)
            self.connect(self.deleteSheetActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.deleteSheetActionTriggered)
        return self.deleteSheetActionVar

    def showNextTabAction(self):
        """ showNextTabAction() -> QAction
        Return the 'Next Sheet' action

        """
        if not hasattr(self, 'showNextTabActionVar'):
            icon = QtGui.QIcon(':/images/forward.png')
            self.showNextTabActionVar = QtGui.QAction(icon, '&Next sheet', self)
            self.showNextTabActionVar.setToolTip('Show the next sheet')
            self.showNextTabActionVar.setStatusTip('Show the next sheet if it '
                                                   'is available')
            self.showNextTabActionShortcut = QtGui.QShortcut(self)
            self.showNextTabActionVar.setShortcut('Ctrl+PgDown')
            self.connect(self.showNextTabActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.showNextTab)
        return self.showNextTabActionVar

    def showPrevTabAction(self):
        """ showPrevTabAction() -> QAction
        Return the 'Prev Sheet' action

        """
        if not hasattr(self, 'showPrevTabActionVar'):
            icon = QtGui.QIcon(':/images/back.png')
            self.showPrevTabActionVar = QtGui.QAction(icon, '&Prev sheet', self)
            self.showPrevTabActionVar.setToolTip('Show the previous sheet')
            self.showPrevTabActionVar.setStatusTip('Show the previous sheet if '
                                                   'it is available')
            self.showPrevTabActionVar.setShortcut('Ctrl+PgUp')
            self.connect(self.showPrevTabActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.showPrevTab)
        return self.showPrevTabActionVar

    def saveAction(self):
        """ saveAction() -> QAction
        Return the 'Save' action

        """
        if not hasattr(self, 'saveActionVar'):
            self.saveActionVar = QtGui.QAction(QtGui.QIcon(':/images/save.png'),
                                               '&Save', self)
            self.saveActionVar.setStatusTip('Save the current spreadsheet')
            self.saveActionVar.setShortcut('Ctrl+S')
            self.connect(self.saveActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.saveSpreadsheet)
        return self.saveActionVar

    def saveAsAction(self):
        """ saveAsAction() -> QAction
        Return the 'Save As...' action

        """
        if not hasattr(self, 'saveAsActionVar'):
            icon = QtGui.QIcon(':/images/saveas.png')
            self.saveAsActionVar = QtGui.QAction(icon, 'Save &As...', self)
            self.saveAsActionVar.setStatusTip('Save the current spreadsheet '
                                              'at a new location')
            self.connect(self.saveAsActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.saveSpreadsheetAs)
        return self.saveAsActionVar

    def openAction(self):
        """ openAction() -> QAction
        Return the 'Open...' action

        """
        if not hasattr(self, 'openActionVar'):
            self.openActionVar = QtGui.QAction(QtGui.QIcon(':/images/open.png'),
                                               '&Open...', self)
            self.openActionVar.setStatusTip('Open a saved spreadsheet')
            self.openActionVar.setShortcut('Ctrl+O')
            self.connect(self.openActionVar,
                         QtCore.SIGNAL('triggered()'),
                         self.openSpreadsheetAs)
        return self.openActionVar

    def exportSheetToImageAction(self):
        """ exportSheetToImageAction() -> QAction
        Export the current sheet to an image

        """
        if not hasattr(self, 'exportSheetToImageVar'):
            self.exportSheetToImageVar = QtGui.QAction('Export Sheet', self)
            self.exportSheetToImageVar.setStatusTip(
                'Export all cells in the spreadsheet to a montaged image')

            exportMenu = QtGui.QMenu(self)
            singleAction = exportMenu.addAction('As a Single Image')
            multiAction = exportMenu.addAction('Separately')
            self.exportSheetToImageVar.setMenu(exportMenu)

            self.connect(self.exportSheetToImageVar,
                         QtCore.SIGNAL('triggered(bool)'),
                         self.exportSheetToSingleImageActionTriggered)

            self.connect(singleAction,
                         QtCore.SIGNAL('triggered()'),
                         self.exportSheetToSingleImageActionTriggered)
            self.connect(multiAction,
                         QtCore.SIGNAL('triggered()'),
                         self.exportSheetToSeparateImagesActionTriggered)
        return self.exportSheetToImageVar

    def exportSheetToSingleImageActionTriggered(self, action=None):
        """ exportSheetToSingleImageActionTriggered() -> None
        Exports the sheet as a big image
        """
        filename = QtGui.QFileDialog.getSaveFileName(
            self, "Select a File to Export the Sheet",
            ".", "Images (*.png *.xpm *.jpg)")
        if filename:
            self.currentWidget().exportSheetToImage(filename)

    def exportSheetToSeparateImagesActionTriggered(self, action=None):
        """ exportSheetToSeparateImagesActionTriggered() -> None
        Exports the cells as separate images
        """
        dirname = QtGui.QFileDialog.getExistingDirectory(
            self, 'Select a Directory to Export Images', ".",
            QtGui.QFileDialog.ShowDirsOnly)
        if dirname:
            self.currentWidget().exportSheetToImages(dirname)

    def newSheetActionTriggered(self, checked=False):
        """ newSheetActionTriggered(checked: boolean) -> None
        Actual code to create a new sheet

        """
        self.setCurrentIndex(self.addTabWidget(StandardWidgetSheetTab(self),
                                               'Sheet %d' % (self.count()+1)))
        self.currentWidget().sheet.stretchCells()

    def tabInserted(self, index):
        """tabInserted(index: int) -> None
        event handler to get when sheets are inserted """
        self.deleteSheetAction().setEnabled(True)
        self.saveAction().setEnabled(True)
        self.saveAsAction().setEnabled(True)

    def tabRemoved(self, index):
        """tabInserted(index: int) -> None
        event handler to get when sheets are removed """
        if self.count() == 0:
            self.deleteSheetAction().setEnabled(False)
            self.saveAction().setEnabled(False)
            self.saveAsAction().setEnabled(False)

    def removeSheetReference(self, sheet):
        """ removeSheetReference(sheet: StandardWidgetSheetTab) -> None
        Remove references of a sheet from the spreadsheet
        """
        for (code, locations) in self.monitoredPipelines.iteritems():
            for lid in reversed(xrange(len(locations))):
                if sheet==locations[lid][0]:
                    del locations[lid]

    def deleteSheetActionTriggered(self, checked=False):
        """ deleteSheetActionTriggered(checked: boolean) -> None
        Actual code to delete the current sheet

        """
        if self.count()>0:
            widget = self.currentWidget()
            self.tabWidgets.remove(widget)
            self.removeTab(self.currentIndex())
            self.removeSheetReference(widget)
            widget.deleteAllCells()
            widget.deleteLater()
            QtCore.QCoreApplication.processEvents()
            gc.collect()

    def clearTabs(self):
        """ clearTabs() -> None
        Clear and reset the controller

        """
        self.executedPipelines = [[], {}, {}]
        while self.count()>0:
            self.deleteSheetActionTriggered()
        for i in reversed(range(len(self.tabWidgets))):
            t = self.tabWidgets[i]
            del self.tabWidgets[i]
            self.removeSheetReference(t)
            t.deleteAllCells()
            t.deleteLater()

    def insertTab(self, idx, tabWidget, tabText):
        """ insertTab(idx: int, tabWidget: QWidget, tabText: str)
                      -> QTabWidget
        Redirect insertTab command to operatingWidget, this can either be a
        QTabWidget or a QStackedWidget

        """
        if self.operatingWidget!=self:
            ret = self.operatingWidget.insertWidget(idx, tabWidget)
            self.operatingWidget.setCurrentIndex(ret)
            return ret
        else:
            return QtGui.QTabWidget.insertTab(self, idx, tabWidget, tabText)

    def findSheet(self, sheetReference):
        """ findSheet(sheetReference: subclass(SheetReference)) -> Sheet widget
        Find/Create a sheet that meets a certen sheet reference

        """
        if not sheetReference:
            return None
        sheetReference.clearCandidate()
        for idx in xrange(len(self.tabWidgets)):
            tabWidget = self.tabWidgets[idx]
            tabLabel = tabWidget.windowTitle()
            sheetReference.checkCandidate(tabWidget, tabLabel, idx,
                                          self.operatingWidget.currentIndex())
        return sheetReference.setupCandidate(self)

    def changeTabText(self, tabIdx, newTabText):
        """ changeTabText(tabIdx: int, newTabText: str) -> None
        Update window title on the operating widget when the tab text
        has changed

        """
        self.operatingWidget.widget(tabIdx).setWindowTitle(newTabText)

    def moveTab(self, tabIdx, destination):
        """ moveTab(tabIdx: int, destination: int) -> None
        Move a tab at tabIdx to a different position at destination

        """
        if (tabIdx<0 or tabIdx>self.count() or
            destination<0 or destination>self.count()):
            return
        tabText = self.tabText(tabIdx)
        tabWidget = self.widget(tabIdx)
        self.removeTab(tabIdx)
        self.insertTab(destination, tabWidget, tabText)
        if tabIdx==self.currentIndex():
            self.setCurrentIndex(destination)

    def splitTab(self, tabIdx, pos=None):
        """ splitTab(tabIdx: int, pos: QPoint) -> None
        Split a tab to be  a stand alone window and move to position pos

        """
        if tabIdx<0 or tabIdx>self.count() or self.count()==0:
            return
        tabWidget = self.widget(tabIdx)
        self.removeTab(tabIdx)

        frame = StandardTabDockWidget(tabWidget.windowTitle(), tabWidget,
                                      self.tabBar(), self)
        if pos:
            frame.move(pos)
        frame.show()
        self.floatingTabWidgets.append(frame)

    def mergeTab(self, frame, tabIdx):
        """ mergeTab(frame: StandardTabDockWidget, tabIdx: int) -> None
        Merge a tab dock widget back to the controller at position tabIdx

        """
        if tabIdx<0 or tabIdx>self.count():
            return
        if tabIdx==self.count(): tabIdx = -1
        tabWidget = frame.widget()
        frame.setWidget(None)
        while frame in self.floatingTabWidgets:
            self.floatingTabWidgets.remove(frame)
        frame.deleteLater()
        tabWidget.setParent(None)
        newIdx = self.insertTab(tabIdx, tabWidget, tabWidget.windowTitle())
        self.setCurrentIndex(newIdx)

    def addTabWidget(self, tabWidget, sheetLabel):
        """ addTabWidget(tabWidget: QWidget, sheetLabel: str) -> int
        Add a new tab widget to the controller

        """
        return self.insertTabWidget(-1, tabWidget, sheetLabel)

    def insertTabWidget(self, index, tabWidget, sheetLabel):
        """ insertTabWidget(index: int, tabWidget: QWidget, sheetLabel: str)
                            -> int
        Insert a tab widget to the controller at some location

        """
        if sheetLabel==None:
            sheetLabel = 'Sheet %d' % (len(self.tabWidgets)+1)
        if not tabWidget in self.tabWidgets:
            self.tabWidgets.append(tabWidget)
            tabWidget.setWindowTitle(sheetLabel)
        return self.insertTab(index, tabWidget, sheetLabel)

    def tabWidgetUnderMouse(self):
        """ tabWidgetUnderMouse() -> QWidget
        Return the tab widget that is under mouse, hide helpers for the rest

        """
        result = None
        for t in self.tabWidgets:
            if t.underMouse():
                result = t
            else:
                t.showHelpers(False, QtCore.QPoint(-1,-1))
        return result

    def setupFullScreenWidget(self, fs, stackedWidget):
        """ setupFullScreenWidget(fs: boolean, stackedWidget: QStackedWidget)
                                  -> None
        Prepare(fs=True)/Clean(fs=False) up full screen mode

        """
        if fs:
            idx = self.currentIndex()
            for i in xrange(self.count()):
                widget = self.widget(0)
                self.removeTab(0)
                self.tabTrueParent = widget.parent()
                widget.setParent(stackedWidget)
                stackedWidget.addWidget(widget)
            stackedWidget.setCurrentIndex(idx)
            self.operatingWidget = stackedWidget
        else:
            idx = stackedWidget.currentIndex()
            for i in xrange(stackedWidget.count()):
                widget = stackedWidget.widget(0)
                stackedWidget.removeWidget(widget)
                widget.setParent(self.tabTrueParent)
                self.addTab(widget, widget.windowTitle())
            self.setCurrentIndex(idx)
            self.operatingWidget = self

    def showNextTab(self):
        """ showNextTab() -> None
        Bring the next tab up

        """
        if self.operatingWidget.currentIndex()<self.operatingWidget.count()-1:
            index = self.operatingWidget.currentIndex()+1
            self.operatingWidget.setCurrentIndex(index)

    def showPrevTab(self):
        """ showPrevTab() -> None
        Bring the previous tab up

        """
        if self.operatingWidget.currentIndex()>0:
            index = self.operatingWidget.currentIndex()-1
            self.operatingWidget.setCurrentIndex(index)

    def tabPopupMenu(self):
        """ tabPopupMenu() -> QMenu
        Return a menu containing a list of all tabs

        """
        menu = QtGui.QMenu(self)
        en = self.operatingWidget.currentIndex()<self.operatingWidget.count()-1
        self.showNextTabAction().setEnabled(en)
        menu.addAction(self.showNextTabAction())
        en = self.operatingWidget.currentIndex()>0
        self.showPrevTabAction().setEnabled(en)
        menu.addAction(self.showPrevTabAction())
        menu.addSeparator()
        for idx in xrange(self.operatingWidget.count()):
            t = self.operatingWidget.widget(idx)
            action = menu.addAction(t.windowTitle())
            action.setData(idx)
            if t==self.operatingWidget.currentWidget():
                action.setIcon(QtGui.QIcon(':/images/ok.png'))
        menu.addAction(self.parent().parent().fullScreenAction())
        return menu

    def showPopupMenu(self):
        """ showPopupMenu() -> None
        Activate the tab list and show the popup menu

        """
        menu = self.tabPopupMenu()
        action = menu.exec_(QtGui.QCursor.pos())
        self.showNextTabAction().setEnabled(True)
        self.showPrevTabAction().setEnabled(True)
        if not action: return
        if not action in self.actions():
            self.operatingWidget.setCurrentIndex(action.data()[0])
        menu.deleteLater()

    def changeSpreadsheetFileName(self, fileName):
        """ changeSpreadsheetFileName(fileName: str) -> None
        Change the current spreadsheet filename and reflect it on the
        window title

        """
        self.spreadsheetFileName = fileName
        if self.spreadsheetFileName:
            displayName = self.spreadsheetFileName
        else:
            displayName = 'Untitled'
        self.emit(QtCore.SIGNAL('needChangeTitle'),
                  '%s - VisTrails Spreadsheet' % displayName)

    def pipelineId(self, pipelineInfo):
        return (pipelineInfo['controller'], pipelineInfo['version'])

    def addPipeline(self, pipelineInfo):
        """ addPipeline(pipelineInfo: dict) -> None
        Add vistrail pipeline executions to history

        """
        vistrail = self.pipelineId(pipelineInfo)
        self.executedPipelines[0].append(vistrail)
        if not vistrail in self.executedPipelines[1]:
            self.executedPipelines[1][vistrail] = 0
        else:
            self.executedPipelines[1][vistrail] += 1
        self.executedPipelines[2][vistrail] = 0

    def getCurrentPipelineId(self, pipelineInfo):
        """ getCurrentPipelineId(pipelineInfo: dict) -> Int
        Get the current pipeline id

        """
        vistrail = self.pipelineId(pipelineInfo)
        return self.executedPipelines[1][vistrail]

    def increasePipelineCellId(self, pipelineInfo):
        """ increasePipelineCellId(pipelineInfo: dict) -> int
        Increase the current cell pipeline id

        """
        vistrail = self.pipelineId(pipelineInfo)
        cid = self.executedPipelines[2][vistrail]
        self.executedPipelines[2][vistrail] += 1
        return cid

    def getCurrentPipelineCellId(self, pipelineInfo):
        """ getCurrentPipelineCellId(pipelineInfo: dict) -> int
        Get current pipeline cell id

        """
        vistrail = self.pipelineId(pipelineInfo)
        return self.executedPipelines[2][vistrail]

    def addPipelineCell(self, pipelineInfo):
        """ addPipelineCell(pipelineInfo: dict) -> None
        Add vistrail pipeline executions to history

        """
        vistrail = self.pipelineId(pipelineInfo)
        self.executedPipelines[0].append(vistrail)
        if not vistrail in self.executedPipelines[1]:
            self.executedPipelines[1][vistrail] = 0
        else:
            self.executedPipelines[1][vistrail] += 1


    def saveSpreadsheet(self, fileName=None):
        """ saveSpreadsheet(fileName: str) -> None
        Save the current spreadsheet to a file if fileName is not
        None. Else, pop up a dialog to ask for a file name.

        """
        def serialize_locator(locator):
            wrapper = XMLWrapper()
            dom = wrapper.create_document('spreadsheet_locator')
            root = dom.documentElement
            root.setAttribute("version", "1.0")
            locator.serialize(dom,root)
            return dom.toxml()

        def need_save():
            from vistrails.gui.vistrails_window import _app
            need_save_vt = False
            for t in self.tabWidgets:
                dim = t.getDimension()
                for r in xrange(dim[0]):
                    for c in xrange(dim[1]):
                        info = t.getCellPipelineInfo(r,c)
                        if info:
                            controller = info[0]['controller']
                            view = _app.ensureController(controller)
                            if view:
                                controller = view.get_controller()
                                if controller.changed:
                                    need_save_vt = True
            return need_save_vt

        if need_save():
            show_warning('Save Spreadsheet', 'Please save your vistrails and try again.')
            return

        if fileName==None:
            fileName = self.spreadsheetFileName
        if fileName:
            indexFile = open(fileName, 'w')
            indexFile.write(str(len(self.tabWidgets))+'\n')
            for t in self.tabWidgets:
                dim = t.getDimension()
                sheet = spreadsheetRegistry.getSheetByType(type(t))
                indexFile.write('%s\n'%str((str(t.windowTitle()),
                                            sheet,
                                            dim[0], dim[1])))
                for r in xrange(dim[0]):
                    for c in xrange(dim[1]):
                        info = t.getCellPipelineInfo(r,c)
                        if info:
                            newinfo0 = copy.copy(info[0])
                            newinfo0['pipeline'] = None
                            newinfo0['actions'] = []
                            newinfo0['locator'] = \
                                          serialize_locator(newinfo0['controller'].locator)
                            newinfo0['controller'] = None
                            indexFile.write('%s\n'
                                            %str((r, c,
                                                  newinfo0,
                                                  info[1], info[2])))
                indexFile.write('---\n')
            indexFile.write(str(len(self.executedPipelines[0]))+'\n')
            for vistrail in self.executedPipelines[0]:
                indexFile.write('%s\n'%str((serialize_locator(vistrail[0].locator),
                                            vistrail[1])))
            self.changeSpreadsheetFileName(fileName)
            indexFile.close()
        else:
            self.saveSpreadsheetAs()

    def saveSpreadsheetAs(self):
        """ saveSpreadsheetAs() -> None
        Asking a file name before saving the spreadsheet

        """
        fileName = QtGui.QFileDialog.getSaveFileName(self,
                                                     'Choose a spreadsheet '
                                                     'name',
                                                     '',
                                                     'VisTrails Spreadsheet '
                                                     '(*.vss)')
        if fileName:
            (root, ext) = os.path.splitext(fileName)
            if ext=='':
                fileName += '.vss'
            self.saveSpreadsheet(fileName)

    def openSpreadsheet(self, fileName):
        """ openSpreadsheet(fileName: str) -> None
        Open a saved spreadsheet assuming that all VTK files must exist and have
        all the version using the saved spreadsheet

        """
        def parse_locator(text):
            locator = None
            wrapper = XMLWrapper()
            dom = wrapper.create_document_from_string(text)
            root = dom.documentElement
            version = None
            version = root.getAttribute('version')
            if version == '1.0':
                for element in named_elements(root, 'locator'):
                    if str(element.getAttribute('type')) == 'file':
                        locator = FileLocator.parse(element)
                    elif str(element.getAttribute('type')) == 'db':
                        locator = DBLocator.parse(element)
            return locator
        locators = {}
        indexFile = open(fileName, 'r')
        contents = indexFile.read()
        self.clearTabs()
        lidx = 0
        lines = contents.split('\n')
        tabCount = int(lines[lidx])
        lidx += 1
        for tabIdx in xrange(tabCount):
            # FIXME: eval should pretty much never be used
            tabInfo = literal_eval(lines[lidx])
            lidx += 1
            sheet = spreadsheetRegistry.getSheet(tabInfo[1])(self)
            sheet.setDimension(tabInfo[2], tabInfo[3])
            self.addTabWidget(sheet, tabInfo[0])
            while lines[lidx]!='---':
                (r, c, vistrail, pid, cid) = literal_eval(lines[lidx])
                locator = vistrail['locator']
                if locators.has_key(locator):
                    vistrail['locator'] = locators[locator]
                else:
                    locators[locator] = parse_locator(vistrail['locator'])
                    vistrail['locator'] = locators[locator]
                self.appendMonitoredLocations((vistrail, pid, cid),
                                              (sheet, r, c))
                lidx += 1
            lidx += 1
        pipelineCount = int(lines[lidx])
        lidx += 1
        self.loadingMode = True
        progress = QtGui.QProgressDialog("Loading spreadsheet...",
                                         "&Cancel", 0, pipelineCount,
                                         self,
                                         QtCore.Qt.WindowStaysOnTopHint
                                         )
        progress.show()
        for pipelineIdx in xrange(pipelineCount):
            # FIXME: eval should pretty much never be used
            (serializedLocator, version) = literal_eval(lines[lidx])
            try:
                locator = locators[serializedLocator]
            except KeyError:
                locator = parse_locator(serializedLocator)
            if locator:
                bundle = locator.load()
                if isinstance(bundle, SaveBundle):
                    pipeline = bundle.vistrail.getPipeline(version)
                else:
                    pipeline = bundle.getPipeline(version)
                execution = get_default_interpreter()
                progress.setValue(pipelineIdx)
                QtCore.QCoreApplication.processEvents()
                if progress.wasCanceled():
                    break
                kwargs = {'locator': locator,
                          'current_version': version,
                          'view': DummyView(),
                          }
                execution.execute(pipeline, **kwargs)
            else:
                raise RuntimeError("Couldn't load spreadsheet")
            lidx += 1
        progress.setValue(pipelineCount)
        QtCore.QCoreApplication.processEvents()
        self.changeSpreadsheetFileName(fileName)
        self.loadingMode = False
        indexFile.close()

    def openSpreadsheetAs(self):
        """ openSpreadsheetAs() -> None
        Open a saved spreadsheet and set its filename in the dialog box

        """
        fileName = QtGui.QFileDialog.getOpenFileName(self,
                                                     'Choose a spreadsheet',
                                                     '',
                                                     'VisTrails Spreadsheet '
                                                     '(*.vss)',
                                                     )
        if fileName:
            self.openSpreadsheet(fileName)

    def cleanup(self):
        """ cleanup() -> None
        Clear reference of non-collectable objects/temp files for gc

        """
        self.clearTabs()

    def setEditingMode(self, editing=True):
        """ setEditingMode(editing: bool) -> None
        Turn on/off the editing mode of the whole spreadsheet

        """
        self.editingMode = editing
        for w in self.tabWidgets:
            w.setEditingMode(editing)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""generated automatically by auto_dao.py"""
from __future__ import division

from vistrails.db.versions.v0_3_0.persistence.xml.xml_dao import XMLDAO
from vistrails.db.versions.v0_3_0.domain.auto_gen import *

class DBChangeParameterXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'set':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        alias = self.convertFromStr(self.getAttribute(node, 'alias'), 'str')
        functionId = self.convertFromStr(self.getAttribute(node, 'functionId'), 'long')
        function = self.convertFromStr(self.getAttribute(node, 'function'), 'str')
        parameterId = self.convertFromStr(self.getAttribute(node, 'parameterId'), 'long')
        parameter = self.convertFromStr(self.getAttribute(node, 'parameter'), 'str')
        type = self.convertFromStr(self.getAttribute(node, 'type'), 'str')
        value = self.convertFromStr(self.getAttribute(node, 'value'), 'str')
        
        return DBChangeParameter(moduleId=moduleId,
                                 alias=alias,
                                 functionId=functionId,
                                 function=function,
                                 parameterId=parameterId,
                                 parameter=parameter,
                                 type=type,
                                 value=value)
    
    def toXML(self, changeParameter, doc):
        node = doc.createElement('set')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(changeParameter.db_moduleId, 'long'))
        node.setAttribute('alias',self.convertToStr(changeParameter.db_alias, 'str'))
        node.setAttribute('functionId',self.convertToStr(changeParameter.db_functionId, 'long'))
        node.setAttribute('function',self.convertToStr(changeParameter.db_function, 'str'))
        node.setAttribute('parameterId',self.convertToStr(changeParameter.db_parameterId, 'long'))
        node.setAttribute('parameter',self.convertToStr(changeParameter.db_parameter, 'str'))
        node.setAttribute('type',self.convertToStr(changeParameter.db_type, 'str'))
        node.setAttribute('value',self.convertToStr(changeParameter.db_value, 'str'))
        
        return node

class DBDeleteFunctionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'function':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        functionId = self.convertFromStr(self.getAttribute(node, 'functionId'), 'long')
        
        return DBDeleteFunction(moduleId=moduleId,
                                functionId=functionId)
    
    def toXML(self, deleteFunction, doc):
        node = doc.createElement('function')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(deleteFunction.db_moduleId, 'long'))
        node.setAttribute('functionId',self.convertToStr(deleteFunction.db_functionId, 'long'))
        
        return node

class DBDeleteConnectionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'connection':
            return None
        
        # read attributes
        connectionId = self.convertFromStr(self.getAttribute(node, 'connectionId'), 'long')
        
        return DBDeleteConnection(connectionId=connectionId)
    
    def toXML(self, deleteConnection, doc):
        node = doc.createElement('connection')
        
        # set attributes
        node.setAttribute('connectionId',self.convertToStr(deleteConnection.db_connectionId, 'long'))
        
        return node

class DBAddModuleXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'object':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        cache = self.convertFromStr(self.getAttribute(node, 'cache'), 'int')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        x = self.convertFromStr(self.getAttribute(node, 'x'), 'float')
        y = self.convertFromStr(self.getAttribute(node, 'y'), 'float')
        
        return DBAddModule(id=id,
                           cache=cache,
                           name=name,
                           x=x,
                           y=y)
    
    def toXML(self, addModule, doc):
        node = doc.createElement('object')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(addModule.db_id, 'long'))
        node.setAttribute('cache',self.convertToStr(addModule.db_cache, 'int'))
        node.setAttribute('name',self.convertToStr(addModule.db_name, 'str'))
        node.setAttribute('x',self.convertToStr(addModule.db_x, 'float'))
        node.setAttribute('y',self.convertToStr(addModule.db_y, 'float'))
        
        return node

class DBDeleteAnnotationXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'annotation':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        key = self.convertFromStr(self.getAttribute(node, 'key'), 'str')
        
        return DBDeleteAnnotation(moduleId=moduleId,
                                  key=key)
    
    def toXML(self, deleteAnnotation, doc):
        node = doc.createElement('annotation')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(deleteAnnotation.db_moduleId, 'long'))
        node.setAttribute('key',self.convertToStr(deleteAnnotation.db_key, 'str'))
        
        return node

class DBDeleteModulePortXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'deletePort':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        portType = self.convertFromStr(self.getAttribute(node, 'portType'), 'str')
        portName = self.convertFromStr(self.getAttribute(node, 'portName'), 'str')
        
        return DBDeleteModulePort(moduleId=moduleId,
                                  portType=portType,
                                  portName=portName)
    
    def toXML(self, deleteModulePort, doc):
        node = doc.createElement('deletePort')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(deleteModulePort.db_moduleId, 'long'))
        node.setAttribute('portType',self.convertToStr(deleteModulePort.db_portType, 'str'))
        node.setAttribute('portName',self.convertToStr(deleteModulePort.db_portName, 'str'))
        
        return node

class DBDeleteModuleXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'module':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        
        return DBDeleteModule(moduleId=moduleId)
    
    def toXML(self, deleteModule, doc):
        node = doc.createElement('module')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(deleteModule.db_moduleId, 'long'))
        
        return node

class DBTagXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'tag':
            return None
        
        # read attributes
        time = self.convertFromStr(self.getAttribute(node, 'time'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        return DBTag(time=time,
                     name=name)
    
    def toXML(self, tag, doc):
        node = doc.createElement('tag')
        
        # set attributes
        node.setAttribute('time',self.convertToStr(tag.db_time, 'long'))
        node.setAttribute('name',self.convertToStr(tag.db_name, 'str'))
        
        return node

class DBAddModulePortXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'addPort':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        portType = self.convertFromStr(self.getAttribute(node, 'portType'), 'str')
        portName = self.convertFromStr(self.getAttribute(node, 'portName'), 'str')
        portSpec = self.convertFromStr(self.getAttribute(node, 'portSpec'), 'str')
        
        return DBAddModulePort(moduleId=moduleId,
                               portType=portType,
                               portName=portName,
                               portSpec=portSpec)
    
    def toXML(self, addModulePort, doc):
        node = doc.createElement('addPort')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(addModulePort.db_moduleId, 'long'))
        node.setAttribute('portType',self.convertToStr(addModulePort.db_portType, 'str'))
        node.setAttribute('portName',self.convertToStr(addModulePort.db_portName, 'str'))
        node.setAttribute('portSpec',self.convertToStr(addModulePort.db_portSpec, 'str'))
        
        return node

class DBActionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'action':
            return None
        
        datas = []
        
        # read attributes
        time = self.convertFromStr(self.getAttribute(node, 'time'), 'long')
        parent = self.convertFromStr(self.getAttribute(node, 'parent'), 'long')
        user = self.convertFromStr(self.getAttribute(node, 'user'), 'str')
        what = self.convertFromStr(self.getAttribute(node, 'what'), 'str')
        date = self.convertFromStr(self.getAttribute(node, 'date'), 'str')
        
        notes = None
        
        # read children
        for child in node.childNodes:
            if child.nodeName == 'notes':
                notes = self.convertFromStr(child.firstChild.nodeValue,'str')
            elif what == 'addModule' and child.nodeName == 'object':
                data = self.getDao('addModule').fromXML(child)
                datas.append(data)
                what = 'addModule'
            elif what == 'addConnection' and child.nodeName == 'connect':
                data = self.getDao('addConnection').fromXML(child)
                datas.append(data)
                what = 'addConnection'
            elif what == 'changeParameter' and child.nodeName == 'set':
                data = self.getDao('changeParameter').fromXML(child)
                datas.append(data)
                what = 'changeParameter'
            elif what == 'changeAnnotation' and child.nodeName == 'set':
                data = self.getDao('changeAnnotation').fromXML(child)
                datas.append(data)
                what = 'changeAnnotation'
            elif what == 'addModulePort' and child.nodeName == 'addPort':
                data = self.getDao('addModulePort').fromXML(child)
                datas.append(data)
                what = 'addModulePort'
            elif what == 'moveModule' and child.nodeName == 'move':
                data = self.getDao('moveModule').fromXML(child)
                datas.append(data)
                what = 'moveModule'
            elif what == 'deleteModule' and child.nodeName == 'module':
                data = self.getDao('deleteModule').fromXML(child)
                datas.append(data)
                what = 'deleteModule'
            elif what == 'deleteConnection' and child.nodeName == 'connection':
                data = self.getDao('deleteConnection').fromXML(child)
                datas.append(data)
                what = 'deleteConnection'
            elif what == 'deleteFunction' and child.nodeName == 'function':
                data = self.getDao('deleteFunction').fromXML(child)
                datas.append(data)
                what = 'deleteFunction'
            elif what == 'deleteAnnotation' and child.nodeName == 'annotation':
                data = self.getDao('deleteAnnotation').fromXML(child)
                datas.append(data)
                what = 'deleteAnnotation'
            elif what == 'deleteModulePort' and child.nodeName == 'deletePort':
                data = self.getDao('deleteModulePort').fromXML(child)
                datas.append(data)
                what = 'deleteModulePort'
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        return DBAction(time=time,
                        parent=parent,
                        user=user,
                        what=what,
                        date=date,
                        notes=notes,
                        datas=datas)
    
    def toXML(self, action, doc):
        node = doc.createElement('action')
        
        # set attributes
        node.setAttribute('time',self.convertToStr(action.db_time, 'long'))
        node.setAttribute('parent',self.convertToStr(action.db_parent, 'long'))
        node.setAttribute('user',self.convertToStr(action.db_user, 'str'))
        node.setAttribute('what',self.convertToStr(action.db_what, 'str'))
        node.setAttribute('date',self.convertToStr(action.db_date, 'str'))
        
        # set elements
        if action.db_notes is not None:
            child = action.db_notes
            notesNode = doc.createElement('notes')
            notesText = doc.createTextNode(self.convertToStr(child, 'str'))
            notesNode.appendChild(notesText)
            node.appendChild(notesNode)
        datas = action.db_datas
        for data in datas:
            if action.db_what == 'addModule':
                node.appendChild(self.getDao('addModule').toXML(data, doc))
            elif action.db_what == 'addConnection':
                node.appendChild(self.getDao('addConnection').toXML(data, doc))
            elif action.db_what == 'changeParameter':
                node.appendChild(self.getDao('changeParameter').toXML(data, doc))
            elif action.db_what == 'changeAnnotation':
                node.appendChild(self.getDao('changeAnnotation').toXML(data, doc))
            elif action.db_what == 'addModulePort':
                node.appendChild(self.getDao('addModulePort').toXML(data, doc))
            elif action.db_what == 'moveModule':
                node.appendChild(self.getDao('moveModule').toXML(data, doc))
            elif action.db_what == 'deleteModule':
                node.appendChild(self.getDao('deleteModule').toXML(data, doc))
            elif action.db_what == 'deleteConnection':
                node.appendChild(self.getDao('deleteConnection').toXML(data, doc))
            elif action.db_what == 'deleteFunction':
                node.appendChild(self.getDao('deleteFunction').toXML(data, doc))
            elif action.db_what == 'deleteAnnotation':
                node.appendChild(self.getDao('deleteAnnotation').toXML(data, doc))
            elif action.db_what == 'deleteModulePort':
                node.appendChild(self.getDao('deleteModulePort').toXML(data, doc))
        
        return node

class DBAddConnectionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'connect':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        destinationId = self.convertFromStr(self.getAttribute(node, 'destinationId'), 'long')
        destinationModule = self.convertFromStr(self.getAttribute(node, 'destinationModule'), 'str')
        destinationPort = self.convertFromStr(self.getAttribute(node, 'destinationPort'), 'str')
        sourceId = self.convertFromStr(self.getAttribute(node, 'sourceId'), 'long')
        sourceModule = self.convertFromStr(self.getAttribute(node, 'sourceModule'), 'str')
        sourcePort = self.convertFromStr(self.getAttribute(node, 'sourcePort'), 'str')
        
        return DBAddConnection(id=id,
                               destinationId=destinationId,
                               destinationModule=destinationModule,
                               destinationPort=destinationPort,
                               sourceId=sourceId,
                               sourceModule=sourceModule,
                               sourcePort=sourcePort)
    
    def toXML(self, addConnection, doc):
        node = doc.createElement('connect')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(addConnection.db_id, 'long'))
        node.setAttribute('destinationId',self.convertToStr(addConnection.db_destinationId, 'long'))
        node.setAttribute('destinationModule',self.convertToStr(addConnection.db_destinationModule, 'str'))
        node.setAttribute('destinationPort',self.convertToStr(addConnection.db_destinationPort, 'str'))
        node.setAttribute('sourceId',self.convertToStr(addConnection.db_sourceId, 'long'))
        node.setAttribute('sourceModule',self.convertToStr(addConnection.db_sourceModule, 'str'))
        node.setAttribute('sourcePort',self.convertToStr(addConnection.db_sourcePort, 'str'))
        
        return node

class DBMoveModuleXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'move':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        dx = self.convertFromStr(self.getAttribute(node, 'dx'), 'float')
        dy = self.convertFromStr(self.getAttribute(node, 'dy'), 'float')
        
        return DBMoveModule(id=id,
                            dx=dx,
                            dy=dy)
    
    def toXML(self, moveModule, doc):
        node = doc.createElement('move')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(moveModule.db_id, 'long'))
        node.setAttribute('dx',self.convertToStr(moveModule.db_dx, 'float'))
        node.setAttribute('dy',self.convertToStr(moveModule.db_dy, 'float'))
        
        return node

class DBVistrailXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'visTrail':
            return None
        
        actions = {}
        tags = {}
        
        # read attributes
        version = self.convertFromStr(self.getAttribute(node, 'version'), 'str')
        
        
        # read children
        for child in node.childNodes:
            if child.nodeName == 'action':
                action = self.getDao('action').fromXML(child)
                actions[action.db_time] = action
            elif child.nodeName == 'tag':
                tag = self.getDao('tag').fromXML(child)
                tags[tag.db_time] = tag
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        return DBVistrail(version=version,
                          actions=actions,
                          tags=tags)
    
    def toXML(self, vistrail, doc):
        node = doc.createElement('visTrail')
        
        # set attributes
        node.setAttribute('version',self.convertToStr(vistrail.db_version, 'str'))
        
        # set elements
        actions = vistrail.db_actions
        for action in actions.itervalues():
            node.appendChild(self.getDao('action').toXML(action, doc))
        tags = vistrail.db_tags
        for tag in tags.itervalues():
            node.appendChild(self.getDao('tag').toXML(tag, doc))
        
        return node

class DBChangeAnnotationXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'set':
            return None
        
        # read attributes
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        key = self.convertFromStr(self.getAttribute(node, 'key'), 'str')
        value = self.convertFromStr(self.getAttribute(node, 'value'), 'str')
        
        return DBChangeAnnotation(moduleId=moduleId,
                                  key=key,
                                  value=value)
    
    def toXML(self, changeAnnotation, doc):
        node = doc.createElement('set')
        
        # set attributes
        node.setAttribute('moduleId',self.convertToStr(changeAnnotation.db_moduleId, 'long'))
        node.setAttribute('key',self.convertToStr(changeAnnotation.db_key, 'str'))
        node.setAttribute('value',self.convertToStr(changeAnnotation.db_value, 'str'))
        
        return node

"""generated automatically by auto_dao.py"""

class XMLDAOListBase(dict):

    def __init__(self, daos=None):
        if daos is not None:
            dict.update(self, daos)

        if 'changeParameter' not in self:
            self['changeParameter'] = DBChangeParameterXMLDAOBase(self)
        if 'deleteFunction' not in self:
            self['deleteFunction'] = DBDeleteFunctionXMLDAOBase(self)
        if 'deleteConnection' not in self:
            self['deleteConnection'] = DBDeleteConnectionXMLDAOBase(self)
        if 'addModule' not in self:
            self['addModule'] = DBAddModuleXMLDAOBase(self)
        if 'deleteAnnotation' not in self:
            self['deleteAnnotation'] = DBDeleteAnnotationXMLDAOBase(self)
        if 'deleteModulePort' not in self:
            self['deleteModulePort'] = DBDeleteModulePortXMLDAOBase(self)
        if 'deleteModule' not in self:
            self['deleteModule'] = DBDeleteModuleXMLDAOBase(self)
        if 'tag' not in self:
            self['tag'] = DBTagXMLDAOBase(self)
        if 'addModulePort' not in self:
            self['addModulePort'] = DBAddModulePortXMLDAOBase(self)
        if 'action' not in self:
            self['action'] = DBActionXMLDAOBase(self)
        if 'addConnection' not in self:
            self['addConnection'] = DBAddConnectionXMLDAOBase(self)
        if 'moveModule' not in self:
            self['moveModule'] = DBMoveModuleXMLDAOBase(self)
        if 'vistrail' not in self:
            self['vistrail'] = DBVistrailXMLDAOBase(self)
        if 'changeAnnotation' not in self:
            self['changeAnnotation'] = DBChangeAnnotationXMLDAOBase(self)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains the definition of the class ParameterExploration """

from __future__ import division

from xml.sax.saxutils import unescape

import vistrails.core.db.action

from vistrails.db.domain import DBParameterExploration, IdScope
from vistrails.core.paramexplore.function import PEFunction
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.vistrail.module import Module
from vistrails.core.modules.paramexplore import IntegerLinearInterpolator, \
   FloatLinearInterpolator, RGBColorInterpolator, HSVColorInterpolator,\
   UserDefinedFunctionInterpolator

from ast import literal_eval
import unittest
import copy

###############################################################################

class ParameterExploration(DBParameterExploration):
    """ParameterExploration

    """

    def __init__(self, *args, **kwargs):
        DBParameterExploration.__init__(self, *args, **kwargs)

        self.set_defaults()

    def __copy__(self):
        """ __copy__() -> ParameterExploration - Returns a clone of itself """
        return ParameterExploration.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBParameterExploration.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ParameterExploration
        cp.set_defaults(self)
        return cp

    def set_defaults(self, other=None):
        if other is None:
            self.changed = False
        else:
            self.changed = other.changed

    @staticmethod
    def convert(_parameter_exploration):
        _parameter_exploration.__class__ = ParameterExploration

        for function in _parameter_exploration.db_functions:
            PEFunction.convert(function)

        _parameter_exploration.set_defaults()

    ##########################################################################
    # Properties

    id = DBParameterExploration.db_id
    action_id = DBParameterExploration.db_action_id
    user = DBParameterExploration.db_user
    date = DBParameterExploration.db_date
    _dims = DBParameterExploration.db_dims
    _layout = DBParameterExploration.db_layout
    name = DBParameterExploration.db_name
    functions = DBParameterExploration.db_functions
    
    def get_dims(self):
        try:
            return literal_eval(self._dims)
        except Exception:
            return []
    def set_dims(self, d):
        try:
            _dims = repr(d)
        except Exception:
            _dims = []
    dims = property(get_dims, set_dims)

    def get_layout(self):
        try:
            return literal_eval(self._layout)
        except Exception:
            return {}
    def set_layout(self, l):
        try:
            _layout = repr(l)
        except Exception:
            _layout = '{}'
    layout = property(get_layout, set_layout)

    def collectParameterActions(self, pipeline):
        """ collectParameterActions() -> list
        Return a list of action lists corresponding to each dimension
        
        """
        if not pipeline:
            return
        unescape_dict = { "&apos;":"'", '&quot;':'"', '&#xa;':'\n' }
        from vistrails.core.modules.module_registry import get_module_registry
        reg = get_module_registry()
        parameterValues = [[], [], [], []]
        # a list of added functions [(module_id, function_name)] = function
        added_functions = {}
        vistrail_vars = []
        function_actions = []
        tmp_f_id = -1L
        tmp_p_id = -1L
        for i in xrange(len(self.functions)):
            pe_function = self.functions[i]
            module = pipeline.db_get_object(Module.vtType, pe_function.module_id)
            # collect overridden vistrail vars
            if module.is_vistrail_var():
                vistrail_vars.append(module.get_vistrail_var())
            port_spec = reg.get_input_port_spec(module, pe_function.port_name)
            for param in pe_function.parameters:
                port_spec_item = port_spec.port_spec_items[param.pos]
                dim = param.dimension
                if dim not in [0, 1, 2, 3]:
                    continue
                count = self.dims[dim]
                # find interpolator values
                values = []
                text = '%s' % unescape(param.value, unescape_dict)
                if param.interpolator == 'Linear Interpolation':
                    # need to figure out type
                    if port_spec_item.module == "Integer":
                        i_range = literal_eval(text)
                        p_min = int(i_range[0])
                        p_max =int(i_range[1])
                        values = IntegerLinearInterpolator(p_min, p_max,
                                                     count).get_values()
                    if port_spec_item.module == "Float":
                        i_range = literal_eval(text)
                        p_min = float(i_range[0])
                        p_max =float(i_range[1])
                        values = FloatLinearInterpolator(p_min, p_max,
                                                     count).get_values()
                elif param.interpolator == 'RGB Interpolation':
                    i_range = literal_eval(text)
                    p_min = str(i_range[0])
                    p_max =str(i_range[1])
                    values = RGBColorInterpolator(p_min, p_max,
                                                     count).get_values()
                elif param.interpolator == 'HSV Interpolation':
                    i_range = literal_eval(text)
                    p_min = str(i_range[0])
                    p_max =str(i_range[1])
                    values = HSVColorInterpolator(p_min, p_max,
                                                     count).get_values()
                elif param.interpolator == 'List':
                    p_module = port_spec_item.descriptor.module
                    values = [p_module.translate_to_python(m)
                              for m in literal_eval(text)]
                elif param.interpolator == 'User-defined Function':
                    p_module = port_spec_item.descriptor.module
                    values = UserDefinedFunctionInterpolator(p_module,
                            text, count).get_values()
                if not values:
                    return None
                # find parameter or create one
                function = [f for f in module.functions
                            if f.name == port_spec.name]
                if function:
                    function = function[0]
                else:
                    try:
                        function = added_functions[(module.id,port_spec.name)]
                    except KeyError:
                        # add to function list
                        params = []
                        for psi in port_spec.port_spec_items:
                            parameter = ModuleParam(id=tmp_p_id,
                                        pos=psi.pos,
                                        name='<no description>',
                                        val=psi.default,
                                        type=psi.descriptor.sigstring) 
                            params.append(parameter)
                            tmp_p_id -= 1
                        function = ModuleFunction(id=tmp_f_id,
                                                  pos=module.getNumFunctions(),
                                                  name=port_spec.name,
                                                  parameters=params)
                        tmp_f_id -= 1
                        added_functions[(module.id, port_spec.name)]=function 
                        action = vistrails.core.db.action.create_action([('add',
                                                                function,
                                                                module.vtType,
                                                                module.id)])
                        function_actions.append(action)
                parameter = function.params[port_spec_item.pos]
                # find old parameter
                old_param = parameter
                actions = []
                for v in values:
                    desc = port_spec_item.descriptor
                    if not isinstance(v, str):
                        str_value = desc.module.translate_to_string(v)
                    else:
                        str_value = v
                    new_param = ModuleParam(id=tmp_p_id,
                                            pos=old_param.pos,
                                            name=old_param.name,
                                            alias=old_param.alias,
                                            val=str_value,
                                            type=old_param.type)
                    tmp_p_id -= 1
                    action_spec = ('change', old_param, new_param,
                                   function.vtType, function.real_id)
                    action = vistrails.core.db.action.create_action([action_spec])
                    actions.append(action)
                parameterValues[dim].append(actions)
        return [zip(*p) for p in parameterValues], function_actions, vistrail_vars

    def __eq__(self, other):
        """ __eq__(other: ParameterExploration) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.action_id != other.action_id:
            return False
        if self._dims != other._dims:
            return False
        if self._layout != other._layout:
            return False
        if len(self.functions) != len(other.functions):
            return False
        for p,q in zip(self.functions, other.functions):
            if p != q:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

# Testing


class TestParameterExploration(unittest.TestCase):

    def create_pe(self, id_scope=IdScope()):
        pe = ParameterExploration(
                            id=id_scope.getNewId(ParameterExploration.vtType),
                            action_id=6,
                            user='tommy',
                            date='2007-11-23 12:48',
                            dims='[1,2]',
                            layout='{1:"normal"}',
                            name='test-pe',
                            functions=[])
        return pe

    def test_copy(self):        
        id_scope = IdScope()
        pe1 = self.create_pe(id_scope)
        pe2 = copy.copy(pe1)
        self.assertEquals(pe1, pe2)
        self.assertEquals(pe1.id, pe2.id)
        pe3 = pe1.do_copy(True, id_scope, {})
        self.assertEquals(pe1, pe3)
        self.assertNotEquals(pe1.id, pe3.id)

    def testComparisonOperators(self):
        """ Test comparison operators """
        p = self.create_pe()
        q = self.create_pe()
        self.assertEqual(p, q)
        q.action_id = 8
        self.assertNotEqual(p, q)
        q.action_id = 6
        q._dims = '[1,4]'
        self.assertNotEqual(p, q)
        q._dims = '[1,2]'
        q._layout = '{1:"different"}'
        self.assertNotEqual(p, q)
        q._layout = p._layout
        q.functions = [1]
        self.assertNotEqual(p, q)

if __name__ == '__main__':
    unittest.main()
